'''
This module provides support for defining relationships between your Elixir
entities.  Elixir currently supports two syntaxes to do so: the default
`Attribute-based syntax`_ which supports the following types of relationships:
ManyToOne_, OneToMany_, OneToOne_ and ManyToMany_, as well as a
`DSL-based syntax`_ which provides the following statements: belongs_to_,
has_many_, has_one_ and has_and_belongs_to_many_.

======================
Attribute-based syntax
======================

The first argument to all these "normal" relationship classes is the name of
the class (entity) you are relating to.

Following that first mandatory argument, any number of additional keyword
arguments can be specified for advanced behavior. See each relationship type
for a list of their specific keyword arguments. At this point, we'll just note
that all the arguments that are not specifically processed by Elixir, as
mentioned in the documentation below are passed on to the SQLAlchemy
``relation`` function. So, please refer to the `SQLAlchemy relation function's
documentation <http://www.sqlalchemy.org/docs/05/reference/orm/mapping.html
#sqlalchemy.orm.relation>`_ for further detail about which
keyword arguments are supported.

You should keep in mind that the following
keyword arguments are automatically generated by Elixir and should not be used
unless you want to override the value provided by Elixir: ``uselist``,
``remote_side``, ``secondary``, ``primaryjoin`` and ``secondaryjoin``.

Additionally, if you want a bidirectionnal relationship, you should define the
inverse relationship on the other entity explicitly (as opposed to how
SQLAlchemy's backrefs are defined). In non-ambiguous situations, Elixir will
match relationships together automatically. If there are several relationships
of the same type between two entities, Elixir is not able to determine which
relationship is the inverse of which, so you have to disambiguate the
situation by giving the name of the inverse relationship in the ``inverse``
keyword argument.

Here is a detailed explanation of each relation type:

`ManyToOne`
-----------

Describes the child's side of a parent-child relationship.  For example,
a `Pet` object may belong to its owner, who is a `Person`.  This could be
expressed like so:

.. sourcecode:: python

    class Pet(Entity):
        owner = ManyToOne('Person')

Behind the scene, assuming the primary key of the `Person` entity is
an integer column named `id`, the ``ManyToOne`` relationship will
automatically add an integer column named `owner_id` to the entity, with a
foreign key referencing the `id` column of the `Person` entity.

In addition to the keyword arguments inherited from SQLAlchemy's relation
function, ``ManyToOne`` relationships accept the following optional arguments
which will be directed to the created column:

+----------------------+------------------------------------------------------+
| Option Name          | Description                                          |
+======================+======================================================+
| ``colname``          | Specify a custom name for the foreign key column(s). |
|                      | This argument accepts either a single string or a    |
|                      | list of strings. The number of strings passed must   |
|                      | match the number of primary key columns of the target|
|                      | entity. If this argument is not used, the name of the|
|                      | column(s) is generated with the pattern              |
|                      | defined in options.FKCOL_NAMEFORMAT, which is, by    |
|                      | default: "%(relname)s_%(key)s", where relname is the |
|                      | name of the ManyToOne relationship, and 'key' is the |
|                      | name (key) of the primary column in the target       |
|                      | entity. That's with, in the above Pet/owner example, |
|                      | the name of the column would be: "owner_id".         |
+----------------------+------------------------------------------------------+
| ``required``         | Specify whether or not this field can be set to None |
|                      | (left without a value). Defaults to ``False``,       |
|                      | unless the field is a primary key.                   |
+----------------------+------------------------------------------------------+
| ``primary_key``      | Specify whether or not the column(s) created by this |
|                      | relationship should act as a primary_key.            |
|                      | Defaults to ``False``.                               |
+----------------------+------------------------------------------------------+
| ``column_kwargs``    | A dictionary holding any other keyword argument you  |
|                      | might want to pass to the Column.                    |
+----------------------+------------------------------------------------------+
| ``target_column``    | Name (or list of names) of the target column(s).     |
|                      | If this argument is not specified, the target entity |
|                      | primary key column(s) are used.                      |
+----------------------+------------------------------------------------------+

The following optional arguments are also supported to customize the
ForeignKeyConstraint that is created:

+----------------------+------------------------------------------------------+
| Option Name          | Description                                          |
+======================+======================================================+
| ``use_alter``        | If True, SQLAlchemy will add the constraint in a     |
|                      | second SQL statement (as opposed to within the       |
|                      | create table statement). This permits to define      |
|                      | tables with a circular foreign key dependency        |
|                      | between them.                                        |
+----------------------+------------------------------------------------------+
| ``ondelete``         | Value for the foreign key constraint ondelete clause.|
|                      | May be one of: ``cascade``, ``restrict``,            |
|                      | ``set null``, or ``set default``.                    |
+----------------------+------------------------------------------------------+
| ``onupdate``         | Value for the foreign key constraint onupdate clause.|
|                      | May be one of: ``cascade``, ``restrict``,            |
|                      | ``set null``, or ``set default``.                    |
+----------------------+------------------------------------------------------+
| ``constraint_kwargs``| A dictionary holding any other keyword argument you  |
|                      | might want to pass to the Constraint.                |
+----------------------+------------------------------------------------------+

In some cases, you may want to declare the foreign key column explicitly,
instead of letting it be generated automatically.  There are several reasons to
that: it could be because you want to declare it with precise arguments and
using column_kwargs makes your code ugly, or because the name of
your column conflicts with the property name (in which case an error is
thrown).  In those cases, you can use the ``field`` argument to specify an
already-declared field to be used for the foreign key column.

For example, for the Pet example above, if you want the database column
(holding the foreign key) to be called 'owner', one should use the field
parameter to specify the field manually.

.. sourcecode:: python

    class Pet(Entity):
        owner_id = Field(Integer, colname='owner')
        owner = ManyToOne('Person', field=owner_id)

+----------------------+------------------------------------------------------+
| Option Name          | Description                                          |
+======================+======================================================+
| ``field``            | Specify the previously-declared field to be used for |
|                      | the foreign key column. Use of this parameter is     |
|                      | mutually exclusive with the colname and column_kwargs|
|                      | arguments.                                           |
+----------------------+------------------------------------------------------+


Additionally, Elixir supports the belongs_to_ statement as an alternative,
DSL-based, syntax to define ManyToOne_ relationships.


`OneToMany`
-----------

Describes the parent's side of a parent-child relationship when there can be
several children.  For example, a `Person` object has many children, each of
them being a `Person`. This could be expressed like so:

.. sourcecode:: python

    class Person(Entity):
        parent = ManyToOne('Person')
        children = OneToMany('Person')

Note that a ``OneToMany`` relationship **cannot exist** without a
corresponding ``ManyToOne`` relationship in the other way. This is because the
``OneToMany`` relationship needs the foreign key created by the ``ManyToOne``
relationship.

In addition to keyword arguments inherited from SQLAlchemy, ``OneToMany``
relationships accept the following optional (keyword) arguments:

+--------------------+--------------------------------------------------------+
| Option Name        | Description                                            |
+====================+========================================================+
| ``order_by``       | Specify which field(s) should be used to sort the      |
|                    | results given by accessing the relation field.         |
|                    | Note that this sort order is only applied when loading |
|                    | objects from the database. Objects appended to the     |
|                    | collection afterwards are not re-sorted in-memory on   |
|                    | the fly.                                               |
|                    | This argument accepts either a string or a list of     |
|                    | strings, each corresponding to the name of a field in  |
|                    | the target entity. These field names can optionally be |
|                    | prefixed by a minus (for descending order).            |
+--------------------+--------------------------------------------------------+
| ``filter``         | Specify a filter criterion (as a clause element) for   |
|                    | this relationship. This criterion will be ``and_`` ed  |
|                    | with the normal join criterion (primaryjoin) generated |
|                    | by Elixir for the relationship. For example:           |
|                    | boston_addresses =                                     |
|                    | OneToMany('Address', filter=Address.city == 'Boston')  |
+--------------------+--------------------------------------------------------+

Additionally, Elixir supports an alternate, DSL-based, syntax to define
OneToMany_ relationships, with the has_many_ statement.


`OneToOne`
----------

Describes the parent's side of a parent-child relationship when there is only
one child.  For example, a `Car` object has one gear stick, which is
represented as a `GearStick` object. This could be expressed like so:

.. sourcecode:: python

    class Car(Entity):
        gear_stick = OneToOne('GearStick', inverse='car')

    class GearStick(Entity):
        car = ManyToOne('Car')

Note that a ``OneToOne`` relationship **cannot exist** without a corresponding
``ManyToOne`` relationship in the other way. This is because the ``OneToOne``
relationship needs the foreign_key created by the ``ManyToOne`` relationship.

Additionally, Elixir supports an alternate, DSL-based, syntax to define
OneToOne_ relationships, with the has_one_ statement.


`ManyToMany`
------------

Describes a relationship in which one kind of entity can be related to several
objects of the other kind but the objects of that other kind can be related to
several objects of the first kind.  For example, an `Article` can have several
tags, but the same `Tag` can be used on several articles.

.. sourcecode:: python

    class Article(Entity):
        tags = ManyToMany('Tag')

    class Tag(Entity):
        articles = ManyToMany('Article')

Behind the scene, the ``ManyToMany`` relationship will automatically create an
intermediate table to host its data.

Note that you don't necessarily need to define the inverse relationship.  In
our example, even though we want tags to be usable on several articles, we
might not be interested in which articles correspond to a particular tag.  In
that case, we could have omitted the `Tag` side of the relationship.

If your ``ManyToMany`` relationship is self-referencial, the entity
containing it is autoloaded (and you don't intend to specify both the
primaryjoin and secondaryjoin arguments manually), you must specify at least
one of either the ``remote_colname`` or ``local_colname`` argument.

In addition to keyword arguments inherited from SQLAlchemy, ``ManyToMany``
relationships accept the following optional (keyword) arguments:

+--------------------+--------------------------------------------------------+
| Option Name        | Description                                            |
+====================+========================================================+
| ``tablename``      | Specify a custom name for the intermediary table. This |
|                    | can be used both when the tables needs to be created   |
|                    | and when the table is autoloaded/reflected from the    |
|                    | database. If this argument is not used, a name will be |
|                    | automatically generated by Elixir depending on the name|
|                    | of the tables of the two entities of the relationship, |
|                    | the name of the relationship, and, if present, the name|
|                    | of its inverse. Even though this argument is optional, |
|                    | it is wise to use it if you are not sure what are the  |
|                    | exact consequence of using a generated table name.     |
+--------------------+--------------------------------------------------------+
| ``schema``         | Specify a custom schema for the intermediate table.    |
|                    | This can be used both when the tables needs to         |
|                    | be created and when the table is autoloaded/reflected  |
|                    | from the database.                                     |
+--------------------+--------------------------------------------------------+
| ``remote_colname`` | A string or list of strings specifying the names of    |
|                    | the column(s) in the intermediary table which          |
|                    | reference the "remote"/target entity's table.          |
+--------------------+--------------------------------------------------------+
| ``local_colname``  | A string or list of strings specifying the names of    |
|                    | the column(s) in the intermediary table which          |
|                    | reference the "local"/current entity's table.          |
+--------------------+--------------------------------------------------------+
| ``table``          | Use a manually created table. If this argument is      |
|                    | used, Elixir won't generate a table for this           |
|                    | relationship, and use the one given instead.           |
+--------------------+--------------------------------------------------------+
| ``order_by``       | Specify which field(s) should be used to sort the      |
|                    | results given by accessing the relation field.         |
|                    | Note that this sort order is only applied when loading |
|                    | objects from the database. Objects appended to the     |
|                    | collection afterwards are not re-sorted in-memory on   |
|                    | the fly.                                               |
|                    | This argument accepts either a string or a list of     |
|                    | strings, each corresponding to the name of a field in  |
|                    | the target entity. These field names can optionally be |
|                    | prefixed by a minus (for descending order).            |
+----------------------+------------------------------------------------------+
| ``ondelete``       | Value for the foreign key constraint ondelete clause.  |
|                    | May be one of: ``cascade``, ``restrict``,              |
|                    | ``set null``, or ``set default``.                      |
+--------------------+--------------------------------------------------------+
| ``onupdate``       | Value for the foreign key constraint onupdate clause.  |
|                    | May be one of: ``cascade``, ``restrict``,              |
|                    | ``set null``, or ``set default``.                      |
+--------------------+--------------------------------------------------------+
| ``table_kwargs``   | A dictionary holding any other keyword argument you    |
|                    | might want to pass to the underlying Table object.     |
+--------------------+--------------------------------------------------------+
| ``column_format``  | DEPRECATED. Specify an alternate format string for     |
|                    | naming the                                             |
|                    | columns in the mapping table.  The default value is    |
|                    | defined in ``elixir.options.M2MCOL_NAMEFORMAT``.  You  |
|                    | will be passed ``tablename``, ``key``, and ``entity``  |
|                    | as arguments to the format string.                     |
+--------------------+--------------------------------------------------------+


================
DSL-based syntax
================

The following DSL statements provide an alternative way to define relationships
between your entities. The first argument to all those statements is the name
of the relationship, the second is the 'kind' of object you are relating to
(it is usually given using the ``of_kind`` keyword).

`belongs_to`
------------

The ``belongs_to`` statement is the DSL syntax equivalent to the ManyToOne_
relationship. As such, it supports all the same arguments as ManyToOne_
relationships.

.. sourcecode:: python

    class Pet(Entity):
        belongs_to('feeder', of_kind='Person')
        belongs_to('owner', of_kind='Person', colname="owner_id")


`has_many`
----------

The ``has_many`` statement is the DSL syntax equivalent to the OneToMany_
relationship. As such, it supports all the same arguments as OneToMany_
relationships.

.. sourcecode:: python

    class Person(Entity):
        belongs_to('parent', of_kind='Person')
        has_many('children', of_kind='Person')

There is also an alternate form of the ``has_many`` relationship that takes
only two keyword arguments: ``through`` and ``via`` in order to encourage a
richer form of many-to-many relationship that is an alternative to the
``has_and_belongs_to_many`` statement.  Here is an example:

.. sourcecode:: python

    class Person(Entity):
        has_field('name', Unicode)
        has_many('assignments', of_kind='Assignment')
        has_many('projects', through='assignments', via='project')

    class Assignment(Entity):
        has_field('start_date', DateTime)
        belongs_to('person', of_kind='Person')
        belongs_to('project', of_kind='Project')

    class Project(Entity):
        has_field('title', Unicode)
        has_many('assignments', of_kind='Assignment')

In the above example, a `Person` has many `projects` through the `Assignment`
relationship object, via a `project` attribute.


`has_one`
---------

The ``has_one`` statement is the DSL syntax equivalent to the OneToOne_
relationship. As such, it supports all the same arguments as OneToOne_
relationships.

.. sourcecode:: python

    class Car(Entity):
        has_one('gear_stick', of_kind='GearStick', inverse='car')

    class GearStick(Entity):
        belongs_to('car', of_kind='Car')


`has_and_belongs_to_many`
-------------------------

The ``has_and_belongs_to_many`` statement is the DSL syntax equivalent to the
ManyToMany_ relationship. As such, it supports all the same arguments as
ManyToMany_ relationships.

.. sourcecode:: python

    class Article(Entity):
        has_and_belongs_to_many('tags', of_kind='Tag')

    class Tag(Entity):
        has_and_belongs_to_many('articles', of_kind='Article')

'''

import warnings

from sqlalchemy import ForeignKeyConstraint, Column, Table, and_
from sqlalchemy.orm import relation, backref, class_mapper
from sqlalchemy.ext.associationproxy import association_proxy

import options
from elixir.statements import ClassMutator
from elixir.properties import Property
from elixir.entity import EntityMeta, DEBUG

__doc_all__ = []


class Relationship(Property):
    '''
    Base class for relationships.
    '''

    def __init__(self, of_kind, inverse=None, *args, **kwargs):
        super(Relationship, self).__init__()

        self.of_kind = of_kind
        self.inverse_name = inverse

        self._target = None

        self.property = None # sqlalchemy property
        self.backref = None  # sqlalchemy backref

        #TODO: unused for now
        self.args = args
        self.kwargs = kwargs

    def attach(self, entity, name):
        super(Relationship, self).attach(entity, name)
        entity._descriptor.relationships.append(self)

    def create_pk_cols(self):
        self.create_keys(True)

    def create_non_pk_cols(self):
        self.create_keys(False)

    def create_keys(self, pk):
        '''
        Subclasses (ie. concrete relationships) may override this method to
        create foreign keys.
        '''

    def create_properties(self):
        if self.property or self.backref:
            return

        kwargs = self.get_prop_kwargs()
        if 'order_by' in kwargs:
            kwargs['order_by'] = \
                self.target._descriptor.translate_order_by(kwargs['order_by'])

        # transform callable arguments
        for arg in ('primaryjoin', 'secondaryjoin', 'remote_side',
                    'foreign_keys'):
            kwarg = kwargs.get(arg, None)
            if hasattr(kwarg, '__call__'):
                kwargs[arg] = kwarg()

        # viewonly relationships need to create "standalone" relations (ie
        # shouldn't be a backref of another relation).
        if self.inverse and not kwargs.get('viewonly', False):
            # check if the inverse was already processed (and thus has already
            # defined a backref we can use)
            if self.inverse.backref:
                # let the user override the backref argument
                if 'backref' not in kwargs:
                    kwargs['backref'] = self.inverse.backref
            else:
                # SQLAlchemy doesn't like when 'secondary' is both defined on
                # the relation and the backref
                kwargs.pop('secondary', None)

                # define backref for use by the inverse
                self.backref = backref(self.name, **kwargs)
                return

        self.property = relation(self.target, **kwargs)
        self.add_mapper_property(self.name, self.property)

    def target(self):
        if not self._target:
            if isinstance(self.of_kind, basestring):
                collection = self.entity._descriptor.collection
                self._target = collection.resolve(self.of_kind, self.entity)
            else:
                self._target = self.of_kind
        return self._target
    target = property(target)

    def inverse(self):
        if not hasattr(self, '_inverse'):
            if self.inverse_name:
                desc = self.target._descriptor
                inverse = desc.find_relationship(self.inverse_name)
                if inverse is None:
                    raise Exception(
                              "Couldn't find a relationship named '%s' in "
                              "entity '%s' or its parent entities."
                              % (self.inverse_name, self.target.__name__))
                assert self.match_type_of(inverse), \
                    "Relationships '%s' in entity '%s' and '%s' in entity " \
                    "'%s' cannot be inverse of each other because their " \
                    "types do not form a valid combination." % \
                    (self.name, self.entity.__name__,
                     self.inverse_name, self.target.__name__)
            else:
                check_reverse = not self.kwargs.get('viewonly', False)
                if isinstance(self.target, EntityMeta):
                    inverse = self.target._descriptor.get_inverse_relation(
                        self, check_reverse=check_reverse)
                else:
                    inverse = None
            self._inverse = inverse
            if inverse and not self.kwargs.get('viewonly', False):
                inverse._inverse = self

        return self._inverse
    inverse = property(inverse)

    def match_type_of(self, other):
        return False

    def is_inverse(self, other):
        # viewonly relationships are not symmetrical: a viewonly relationship
        # should have exactly one inverse (a ManyToOne relationship), but that
        # inverse shouldn't have the viewonly relationship as its inverse.
        return not other.kwargs.get('viewonly', False) and \
               other is not self and \
               self.match_type_of(other) and \
               self.entity == other.target and \
               other.entity == self.target and \
               (self.inverse_name == other.name or not self.inverse_name) and \
               (other.inverse_name == self.name or not other.inverse_name)


class ManyToOne(Relationship):
    '''

    '''

    def __init__(self, of_kind,
                 column_kwargs=None,
                 colname=None, required=None, primary_key=None,
                 field=None,
                 constraint_kwargs=None,
                 use_alter=None, ondelete=None, onupdate=None,
                 target_column=None,
                 *args, **kwargs):

        # 1) handle column-related args

        # check that the column arguments don't conflict
        assert not (field and (column_kwargs or colname)), \
               "ManyToOne can accept the 'field' argument or column " \
               "arguments ('colname' or 'column_kwargs') but not both!"

        if colname and not isinstance(colname, list):
            colname = [colname]
        self.colname = colname or []

        column_kwargs = column_kwargs or {}
        # kwargs go by default to the relation(), so we need to manually
        # extract those targeting the Column
        if required is not None:
            column_kwargs['nullable'] = not required
        if primary_key is not None:
            column_kwargs['primary_key'] = primary_key
        # by default, created columns will have an index.
        column_kwargs.setdefault('index', True)
        self.column_kwargs = column_kwargs

        if field and not isinstance(field, list):
            field = [field]
        self.field = field or []

        # 2) handle constraint kwargs
        constraint_kwargs = constraint_kwargs or {}
        if use_alter is not None:
            constraint_kwargs['use_alter'] = use_alter
        if ondelete is not None:
            constraint_kwargs['ondelete'] = ondelete
        if onupdate is not None:
            constraint_kwargs['onupdate'] = onupdate
        self.constraint_kwargs = constraint_kwargs

        # 3) misc arguments
        if target_column and not isinstance(target_column, list):
            target_column = [target_column]
        self.target_column = target_column

        self.foreign_key = []
        self.primaryjoin_clauses = []

        super(ManyToOne, self).__init__(of_kind, *args, **kwargs)

    def match_type_of(self, other):
        return isinstance(other, (OneToMany, OneToOne))

    def target_table(self):
        if isinstance(self.target, EntityMeta):
            return self.target._descriptor.table
        else:
            return class_mapper(self.target).local_table
    target_table = property(target_table)

    def create_keys(self, pk):
        '''
        Find all primary keys on the target and create foreign keys on the
        source accordingly.
        '''

        if self.foreign_key:
            return

        if self.column_kwargs.get('primary_key', False) != pk:
            return

        source_desc = self.entity._descriptor
        if isinstance(self.target, EntityMeta):
            # make sure the target has all its pk set up
            self.target._descriptor.create_pk_cols()
        #XXX: another option, instead of the FakeTable, would be to create an
        # EntityDescriptor for the SA class.
        target_table = self.target_table

        if source_desc.autoload:
            #TODO: allow target_column to be used as an alternative to
            # specifying primaryjoin, to be consistent with non-autoloaded
            # tables
            if self.colname:
                if 'primaryjoin' not in self.kwargs:
                    self.primaryjoin_clauses = \
                        _get_join_clauses(self.entity.table,
                                          self.colname, None,
                                          target_table)[0]
                    if not self.primaryjoin_clauses:
                        colnames = ', '.join(self.colname)
                        raise Exception(
                            "Couldn't find a foreign key constraint in table "
                            "'%s' using the following columns: %s."
                            % (self.entity.table.name, colnames))
            if self.field:
                raise NotImplementedError(
                    "'field' argument not allowed on autoloaded table "
                    "relationships.")
        else:
            fk_refcols = []
            fk_colnames = []

            if self.target_column is None:
                target_columns = target_table.primary_key.columns
            else:
                target_columns = [target_table.columns[col]
                                  for col in self.target_column]

            if not target_columns:
                raise Exception("No primary key found in target table ('%s') "
                                "for the '%s' relationship of the '%s' entity."
                                % (target_table.name, self.name,
                                   self.entity.__name__))
            if self.colname and \
               len(self.colname) != len(target_columns):
                raise Exception(
                        "The number of column names provided in the colname "
                        "keyword argument of the '%s' relationship of the "
                        "'%s' entity is not the same as the number of columns "
                        "of the primary key of '%s'."
                        % (self.name, self.entity.__name__,
                           self.target.__name__))

            for key_num, target_col in enumerate(target_columns):
                if self.field:
                    col = self.field[key_num].column
                else:
                    if self.colname:
                        colname = self.colname[key_num]
                    else:
                        colname = options.FKCOL_NAMEFORMAT % \
                                  {'relname': self.name,
                                   'key': target_col.key}

                    # We can't add the column to the table directly as the
                    # table might not be created yet.
                    col = Column(colname, target_col.type,
                                 **self.column_kwargs)
                    source_desc.add_column(col)

                    # If the column name was specified, and it is the same as
                    # this property's name, there is going to be a conflict.
                    # Don't allow this to happen.
                    if col.key == self.name:
                        raise ValueError(
                                 "ManyToOne named '%s' in '%s' conficts "
                                 " with the column of the same name. "
                                 "You should probably define the foreign key "
                                 "field manually and use the 'field' "
                                 "argument on the ManyToOne relationship"
                                 % (self.name, self.entity.__name__))

                # Build the list of local columns which will be part of
                # the foreign key
                self.foreign_key.append(col)

                # Store the names of those columns
                fk_colnames.append(col.key)

                # Build the list of column "paths" the foreign key will
                # point to
                fk_refcols.append("%s.%s" % \
                                  (target_table.fullname, target_col.key))

                # Build up the primary join. This is needed when you have
                # several ManyToOne relationships between two objects
                self.primaryjoin_clauses.append(col == target_col)

            if 'name' not in self.constraint_kwargs:
                # In some databases (at least MySQL) the constraint name needs
                # to be unique for the whole database, instead of per table.
                fk_name = options.CONSTRAINT_NAMEFORMAT % \
                          {'tablename': source_desc.tablename,
                           'colnames': '_'.join(fk_colnames)}
                self.constraint_kwargs['name'] = fk_name

            source_desc.add_constraint(
                ForeignKeyConstraint(fk_colnames, fk_refcols,
                                     **self.constraint_kwargs))

    def get_prop_kwargs(self):
        kwargs = {'uselist': False}

        if self.entity.table is self.target_table:
            # this is needed because otherwise SA has no way to know what is
            # the direction of the relationship since both columns present in
            # the primaryjoin belong to the same table. In other words, it is
            # necessary to know if this particular relation
            # is the many-to-one side, or the one-to-xxx side. The foreignkey
            # doesn't help in this case.
            kwargs['remote_side'] = \
                [col for col in self.target_table.primary_key.columns]

        if self.primaryjoin_clauses:
            kwargs['primaryjoin'] = and_(*self.primaryjoin_clauses)

        kwargs.update(self.kwargs)

        return kwargs


class OneToOne(Relationship):
    uselist = False

    def __init__(self, of_kind, filter=None, *args, **kwargs):
        self.filter = filter
        if filter is not None:
            # We set viewonly to True by default for filtered relationships,
            # unless manually overridden.
            # This is not strictly necessary, as SQLAlchemy allows non viewonly
            # relationships with a custom join/filter. The example at:
            # SADOCS/05/mappers.html#advdatamapping_relation_customjoin
            # is not viewonly. Those relationships can be used as if the extra
            # filter wasn't present when inserting. This can lead to a
            # confusing behavior (if you insert data which doesn't match the
            # extra criterion it'll get inserted anyway but you won't see it
            # when you query back the attribute after a round-trip to the
            # database).
            if 'viewonly' not in kwargs:
                kwargs['viewonly'] = True
        super(OneToOne, self).__init__(of_kind, *args, **kwargs)

    def match_type_of(self, other):
        return isinstance(other, ManyToOne)

    def create_keys(self, pk):
        # make sure an inverse relationship exists
        if self.inverse is None:
            raise Exception(
                      "Couldn't find any relationship in '%s' which "
                      "match as inverse of the '%s' relationship "
                      "defined in the '%s' entity. If you are using "
                      "inheritance you "
                      "might need to specify inverse relationships "
                      "manually by using the 'inverse' argument."
                      % (self.target, self.name,
                         self.entity))

    def get_prop_kwargs(self):
        kwargs = {'uselist': self.uselist}

        #TODO: for now, we don't break any test if we remove those 2 lines.
        # So, we should either complete the selfref test to prove that they
        # are indeed useful, or remove them. It might be they are indeed
        # useless because the remote_side is already setup in the other way
        # (ManyToOne).
        if self.entity.table is self.target.table:
            #FIXME: IF this code is of any use, it will probably break for
            # autoloaded tables
            kwargs['remote_side'] = self.inverse.foreign_key

        # Contrary to ManyToMany relationships, we need to specify the join
        # clauses even if this relationship is not self-referencial because
        # there could be several ManyToOne from the target class to us.
        joinclauses = self.inverse.primaryjoin_clauses
        if self.filter:
            # We need to make a copy of the joinclauses, to not add the filter
            # on the backref
            joinclauses = joinclauses[:] + [self.filter(self.target.table.c)]
        if joinclauses:
            kwargs['primaryjoin'] = and_(*joinclauses)

        kwargs.update(self.kwargs)

        return kwargs


class OneToMany(OneToOne):
    uselist = True


class ManyToMany(Relationship):
    uselist = True

    def __init__(self, of_kind, tablename=None,
                 local_colname=None, remote_colname=None,
                 ondelete=None, onupdate=None,
                 table=None, schema=None,
                 column_format=None,
                 filter=None,
                 table_kwargs=None,
                 *args, **kwargs):
        self.user_tablename = tablename

        if local_colname and not isinstance(local_colname, list):
            local_colname = [local_colname]
        self.local_colname = local_colname or []
        if remote_colname and not isinstance(remote_colname, list):
            remote_colname = [remote_colname]
        self.remote_colname = remote_colname or []

        self.ondelete = ondelete
        self.onupdate = onupdate

        self.table = table
        self.schema = schema

        if column_format:
            warnings.warn("The 'column_format' argument on ManyToMany "
                "relationships is deprecated. Please use the 'local_colname' "
                "and/or 'remote_colname' arguments if you want custom "
                "column names for this table only, or modify "
                "options.M2MCOL_NAMEFORMAT if you want a custom format for "
                "all ManyToMany tables", DeprecationWarning, stacklevel=3)
        self.column_format = column_format or options.M2MCOL_NAMEFORMAT
        if not hasattr(self.column_format, '__call__'):
            # we need to store the format in a variable so that the
            # closure of the lambda is correct
            format = self.column_format
            self.column_format = lambda data: format % data
        if options.MIGRATION_TO_07_AID:
            self.column_format = \
                migration_aid_m2m_column_formatter(
                    lambda data: options.OLD_M2MCOL_NAMEFORMAT % data,
                    self.column_format)

        self.filter = filter
        if filter is not None:
            # We set viewonly to True by default for filtered relationships,
            # unless manually overridden.
            if 'viewonly' not in kwargs:
                kwargs['viewonly'] = True

        self.table_kwargs = table_kwargs or {}

        self.primaryjoin_clauses = []
        self.secondaryjoin_clauses = []

        super(ManyToMany, self).__init__(of_kind, *args, **kwargs)

    def get_table(self):
        warnings.warn("The secondary_table attribute on ManyToMany objects is "
                      "deprecated. You should rather use the table attribute.",
                      DeprecationWarning, stacklevel=2)
        return self.table
    secondary_table = property(get_table)

    def match_type_of(self, other):
        return isinstance(other, ManyToMany)

    def create_tables(self):
        if self.table is not None:
            if 'primaryjoin' not in self.kwargs or \
               'secondaryjoin' not in self.kwargs:
                self._build_join_clauses()
            assert self.inverse is None or self.inverse.table is None or \
                   self.inverse.table is self.table
            return

        if self.inverse:
            inverse = self.inverse
            if inverse.table is not None:
                self.table = inverse.table
                self.primaryjoin_clauses = inverse.secondaryjoin_clauses
                self.secondaryjoin_clauses = inverse.primaryjoin_clauses
                return

            assert not inverse.user_tablename or not self.user_tablename or \
                   inverse.user_tablename == self.user_tablename
            assert not inverse.remote_colname or not self.local_colname or \
                   inverse.remote_colname == self.local_colname
            assert not inverse.local_colname or not self.remote_colname or \
                   inverse.local_colname == self.remote_colname
            assert not inverse.schema or not self.schema or \
                   inverse.schema == self.schema
            assert not inverse.table_kwargs or not self.table_kwargs or \
                   inverse.table_kwargs == self.table_kwargs

            self.user_tablename = inverse.user_tablename or self.user_tablename
            self.local_colname = inverse.remote_colname or self.local_colname
            self.remote_colname = inverse.local_colname or self.remote_colname
            self.schema = inverse.schema or self.schema
            self.local_colname = inverse.remote_colname or self.local_colname

        # compute table_kwargs
        complete_kwargs = options.options_defaults['table_options'].copy()
        complete_kwargs.update(self.table_kwargs)

        #needs: table_options['schema'], autoload, tablename, primary_keys,
        #entity.__name__, table_fullname
        e1_desc = self.entity._descriptor
        e2_desc = self.target._descriptor

        e1_schema = e1_desc.table_options.get('schema', None)
        e2_schema = e2_desc.table_options.get('schema', None)
        schema = (self.schema is not None) and self.schema or e1_schema

        assert e1_schema == e2_schema or self.schema, \
               "Schema %r for entity %s differs from schema %r of entity %s." \
               " Consider using the schema-parameter. "\
               % (e1_schema, self.entity.__name__,
                  e2_schema, self.target.__name__)

        # First, we compute the name of the table. Note that some of the
        # intermediary variables are reused later for the constraint
        # names.

        # We use the name of the relation for the first entity
        # (instead of the name of its primary key), so that we can
        # have two many-to-many relations between the same objects
        # without having a table name collision.
        source_part = "%s_%s" % (e1_desc.tablename, self.name)

        # And we use only the name of the table of the second entity
        # when there is no inverse, so that a many-to-many relation
        # can be defined without an inverse.
        if self.inverse:
            target_part = "%s_%s" % (e2_desc.tablename, self.inverse.name)
        else:
            target_part = e2_desc.tablename

        if self.user_tablename:
            tablename = self.user_tablename
        else:
            # We need to keep the table name consistent (independant of
            # whether this relation or its inverse is setup first).
            if self.inverse and source_part < target_part:
                #XXX: use a different scheme for selfref (to not include the
                #     table name twice)?
                tablename = "%s__%s" % (target_part, source_part)
            else:
                tablename = "%s__%s" % (source_part, target_part)

            if options.MIGRATION_TO_07_AID:
                oldname = (self.inverse and
                           e1_desc.tablename < e2_desc.tablename) and \
                          "%s__%s" % (target_part, source_part) or \
                          "%s__%s" % (source_part, target_part)
                if oldname != tablename:
                    warnings.warn(
                        "The generated table name for the '%s' relationship "
                        "on the '%s' entity changed from '%s' (the name "
                        "generated by Elixir 0.6.1 and earlier) to '%s'. "
                        "You should either rename the table in the database "
                        "to the new name or use the tablename argument on the "
                        "relationship to force the old name: tablename='%s'!"
                        % (self.name, self.entity.__name__, oldname,
                           tablename, oldname))

        if e1_desc.autoload:
            if not e2_desc.autoload:
                raise Exception(
                    "Entity '%s' is autoloaded and its '%s' "
                    "ManyToMany relationship points to "
                    "the '%s' entity which is not autoloaded"
                    % (self.entity.__name__, self.name,
                       self.target.__name__))

            self.table = Table(tablename, e1_desc.metadata, autoload=True,
                               **complete_kwargs)
            if 'primaryjoin' not in self.kwargs or \
               'secondaryjoin' not in self.kwargs:
                self._build_join_clauses()
        else:
            # We pre-compute the names of the foreign key constraints
            # pointing to the source (local) entity's table and to the
            # target's table

            # In some databases (at least MySQL) the constraint names need
            # to be unique for the whole database, instead of per table.
            source_fk_name = "%s_fk" % source_part
            if self.inverse:
                target_fk_name = "%s_fk" % target_part
            else:
                target_fk_name = "%s_inverse_fk" % source_part

            columns = []
            constraints = []

            for num, desc, fk_name, rel, inverse, colnames, join_clauses in (
              (0, e1_desc, source_fk_name, self, self.inverse,
               self.local_colname, self.primaryjoin_clauses),
              (1, e2_desc, target_fk_name, self.inverse, self,
               self.remote_colname, self.secondaryjoin_clauses)):

                fk_colnames = []
                fk_refcols = []
                if colnames:
                    assert len(colnames) == len(desc.primary_keys)
                else:
                    # The data generated here will be fed to the M2M column
                    # formatter to generate the name of the columns of the
                    # intermediate table for *one* side of the relationship,
                    # that is, from the intermediate table to the current
                    # entity, as stored in the "desc" variable.
                    data = {# A) relationships info

                            # the name of the rel going *from* the entity
                            # we are currently generating a column pointing
                            # *to*. This is generally *not* what you want to
                            # use. eg in a "Post" and "Tag" example, with
                            # relationships named 'tags' and 'posts', when
                            # creating the columns from the intermediate
                            # table to the "Post" entity, 'relname' will
                            # contain 'tags'.
                            'relname': rel and rel.name or 'inverse',

                            # the name of the inverse relationship. In the
                            # above example, 'inversename' will contain
                            # 'posts'.
                            'inversename': inverse and inverse.name
                                                   or 'inverse',
                            # is A == B?
                            'selfref': e1_desc is e2_desc,
                            # provided for backward compatibility, DO NOT USE!
                            'num': num,
                            # provided for backward compatibility, DO NOT USE!
                            'numifself': e1_desc is e2_desc and str(num + 1)
                                                            or '',
                            # B) target information (from the perspective of
                            #    the intermediate table)
                            'target': desc.entity,
                            'entity': desc.entity.__name__.lower(),
                            'tablename': desc.tablename,

                            # C) current (intermediate) table name
                            'current_table': tablename
                           }
                    colnames = []
                    for pk_col in desc.primary_keys:
                        data.update(key=pk_col.key)
                        colnames.append(self.column_format(data))

                for pk_col, colname in zip(desc.primary_keys, colnames):
                    col = Column(colname, pk_col.type, primary_key=True)
                    columns.append(col)

                    # Build the list of local columns which will be part
                    # of the foreign key.
                    fk_colnames.append(colname)

                    # Build the list of column "paths" the foreign key will
                    # point to
                    target_path = "%s.%s" % (desc.table_fullname, pk_col.key)
                    fk_refcols.append(target_path)

                    # Build join clauses (in case we have a self-ref)
                    if self.entity is self.target:
                        join_clauses.append(col == pk_col)

                onupdate = rel and rel.onupdate
                ondelete = rel and rel.ondelete

                #FIXME: fk_name is misleading
                constraints.append(
                    ForeignKeyConstraint(fk_colnames, fk_refcols,
                                         name=fk_name, onupdate=onupdate,
                                         ondelete=ondelete))

            args = columns + constraints

            self.table = Table(tablename, e1_desc.metadata,
                               schema=schema, *args, **complete_kwargs)
            if DEBUG:
                print self.table.repr2()

    def _build_join_clauses(self):
        # In the case we have a self-reference, we need to build join clauses
        if self.entity is self.target:
            if not self.local_colname and not self.remote_colname:
                raise Exception(
                    "Self-referential ManyToMany "
                    "relationships in autoloaded entities need to have at "
                    "least one of either 'local_colname' or 'remote_colname' "
                    "argument specified. The '%s' relationship in the '%s' "
                    "entity doesn't have either."
                    % (self.name, self.entity.__name__))

            self.primaryjoin_clauses, self.secondaryjoin_clauses = \
                _get_join_clauses(self.table,
                                  self.local_colname, self.remote_colname,
                                  self.entity.table)

    def get_prop_kwargs(self):
        kwargs = {'secondary': self.table,
                  'uselist': self.uselist}

        if self.filter:
            # we need to make a copy of the joinclauses
            secondaryjoin_clauses = self.secondaryjoin_clauses[:] + \
                                    [self.filter(self.target.table.c)]
        else:
            secondaryjoin_clauses = self.secondaryjoin_clauses

        if self.target is self.entity or self.filter:
            kwargs['primaryjoin'] = and_(*self.primaryjoin_clauses)
            kwargs['secondaryjoin'] = and_(*secondaryjoin_clauses)

        kwargs.update(self.kwargs)

        return kwargs

    def is_inverse(self, other):
        return super(ManyToMany, self).is_inverse(other) and \
               (self.user_tablename == other.user_tablename or
                (not self.user_tablename and not other.user_tablename))


def migration_aid_m2m_column_formatter(oldformatter, newformatter):
    def debug_formatter(data):
        old_name = oldformatter(data)
        new_name = newformatter(data)
        if new_name != old_name:
            complete_data = data.copy()
            complete_data.update(old_name=old_name,
                                 new_name=new_name,
                                 targetname=data['target'].__name__)
            # Specifying a stacklevel is useless in this case as the name
            # generation is triggered by setup_all(), not by the declaration
            # of the offending relationship.
            warnings.warn("The '%(old_name)s' column in the "
                          "'%(current_table)s' table, used as the "
                          "intermediate table for the '%(relname)s' "
                          "relationship on the '%(targetname)s' entity "
                          "was renamed to '%(new_name)s'."
                          % complete_data)
        return new_name
    return debug_formatter


def _get_join_clauses(local_table, local_cols1, local_cols2, target_table):
    primary_join, secondary_join = [], []
    cols1 = local_cols1[:]
    cols1.sort()
    cols1 = tuple(cols1)

    if local_cols2 is not None:
        cols2 = local_cols2[:]
        cols2.sort()
        cols2 = tuple(cols2)
    else:
        cols2 = None

    # Build a map of fk constraints pointing to the correct table.
    # The map is indexed on the local col names.
    constraint_map = {}
    for constraint in local_table.constraints:
        if isinstance(constraint, ForeignKeyConstraint):
            use_constraint = True
            fk_colnames = []

            # if all columns point to the correct table, we use the constraint
            #TODO: check that it contains as many columns as the pk of the
            #target entity, or even that it points to the actual pk columns
            for fk in constraint.elements:
                if fk.references(target_table):
                    # local column key
                    fk_colnames.append(fk.parent.key)
                else:
                    use_constraint = False
            if use_constraint:
                fk_colnames.sort()
                constraint_map[tuple(fk_colnames)] = constraint

    # Either the fk column names match explicitely with the columns given for
    # one of the joins (primary or secondary), or we assume the current
    # columns match because the columns for this join were not given and we
    # know the other join is either not used (is None) or has an explicit
    # match.

#TODO: rewrite this. Even with the comment, I don't even understand it myself.
    for cols, constraint in constraint_map.iteritems():
        if cols == cols1 or (cols != cols2 and
                             not cols1 and (cols2 in constraint_map or
                                            cols2 is None)):
            join = primary_join
        elif cols == cols2 or (cols2 == () and cols1 in constraint_map):
            join = secondary_join
        else:
            continue
        for fk in constraint.elements:
            join.append(fk.parent == fk.column)
    return primary_join, secondary_join


def rel_mutator_handler(target):
    def handler(entity, name, of_kind=None, through=None, via=None,
                *args, **kwargs):
        if through and via:
            setattr(entity, name,
                    association_proxy(through, via, **kwargs))
            return
        elif through or via:
            raise Exception("'through' and 'via' relationship keyword "
                            "arguments should be used in combination.")
        rel = target(of_kind, *args, **kwargs)
        rel.attach(entity, name)
    return handler


belongs_to = ClassMutator(rel_mutator_handler(ManyToOne))
has_one = ClassMutator(rel_mutator_handler(OneToOne))
has_many = ClassMutator(rel_mutator_handler(OneToMany))
has_and_belongs_to_many = ClassMutator(rel_mutator_handler(ManyToMany))



# Copyright (C) 2004,2005 by SICEm S.L. and Imendio AB
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

import gtk

from gazpacho import util

from gazpacho.clipboard import CommandCutPaste
from gazpacho.command import Command, FlipFlopCommandMixin
from gazpacho.gadget import Gadget
from gazpacho.i18n import _
from gazpacho.placeholder import Placeholder
from gazpacho.properties import prop_registry, CustomProperty, IntType
from gazpacho.util import get_parent
from gazpacho.widgets.base.base import ContainerAdaptor
from gazpacho.widgetregistry import widget_registry

class BoxAdaptor(ContainerAdaptor):
    def fill_empty(self, context, widget):
        pass

# GtkBox
class BoxSizeProp(CustomProperty, IntType):
    minimum = 1
    default = 3
    label = 'Size'
    persistent = False

    def __init__(self, object):
        super(BoxSizeProp, self).__init__(object)

    def load(self):
        self._initial = self.default

        # Don't set default if object has childs already
        if not self.get():
            self.set(self.default)

    def get(self):
        return len(self.object.get_children())

    def set(self, new_size):
        old_size = len(self.object.get_children())
        if new_size == old_size:
            return
        elif new_size > old_size:
            # The box has grown. Add placeholders
            while new_size > old_size:
                self.object.add(Placeholder())
                old_size += 1
        elif new_size > 0:
            # The box has shrunk. Remove placeholders first, starting
            # with the last one
            for child in self.object.get_children()[::-1]:
                if isinstance(child, Placeholder):
                    gtk.Container.remove(self.object, child)
                    old_size -= 1

                if old_size == new_size:
                    return

            # and then remove widgets
            child = self.object.get_children()[-1]
            while old_size > new_size and child:
                gadget = Gadget.from_widget(child)
                if gadget: # It may be None, e.g a placeholder
                    gadget.project.remove_widget(child)

                gtk.Container.remove(self.object, child)
                child = self.object.get_children()[-1]
                old_size -= 1
        self.notify()

prop_registry.override_property('GtkBox::size', BoxSizeProp)
prop_registry.override_simple('GtkBox::spacing', minimum=0)

class AppendGadgetCommand(Command):
    """
    Command for adding a widget to a box. This command isn't inteded
    to be used directly but to be subclasses by other commands.
    """

    def __init__(self, source_gadget, box, pos, description=None):
        """
        Initialize the command. Note that the source_gadget has to be
        a new widget that is not already in use in the project.

        @param source_gadget: the widget that is to be inserted
        @type source_gadget: L{gazpacho.gadget.Gadget}
        @param box: the box into which the widget should be inserted
        @type box: L{gazpacho.gadget.Gadget}
        @param pos: the position where the widget will be inserted
        @type pos: int
        """
        Command.__init__(self, description)

        self._box = box.widget
        self._pos = pos
        self._source = source_gadget.widget
        self._project = box.project


    def insert_execute(self):
        """
        Insert the widget into the box.
        """
        self._box.add(self._source)
        self._box.reorder_child(self._source, self._pos)

        self._project.add_widget(self._source)
        self._project.selection.set(self._source)
        self._source.show_all()

    def remove_execute(self):
        """
        Remove the widget from the box.
        """
        self._box.remove(self._source)
        self._source.hide()
        self._project.remove_widget(self._source)

class DragAppendCommand(AppendGadgetCommand):
    """
    Append a copy of an existing widget to a box. It's optional to
    remove the source widget as well.
    """

    def __init__(self, source_gadget, box, pos, keep_source):
        """
        Initialize the command.

        @param source_gadget:
        @type source_gadget: L{gazpacho.gadget.Gadget}
        @param box: the box into which the widget should be inserted
        @type box: L{gazpacho.gadget.Gadget}
        @param pos: the position where the widget will be inserted
        @type pos: int
        """
        description = _("Drag - append")

        # We append a copy of the widget
        source_copy = source_gadget.copy(box.project, not keep_source)
        AppendGadgetCommand.__init__(self, source_copy, box, pos, description)

        self._undo = False

        # Reuse the Cut command for removing the widget
        self._remove_cmd = None
        if not keep_source:
            self._remove_cmd = CommandCutPaste(source_gadget,
                                               source_gadget.project,
                                               None,
                                               True)

    def execute(self):
        """
        Execute the command.
        """
        if self._undo:
            self.remove_execute()

        if self._remove_cmd:
            self._remove_cmd.execute()

        if not self._undo:
            self.insert_execute()

        self._undo = not self._undo

class CreateAppendCommand(AppendGadgetCommand):
    """
    Append a newly created widget to a box.
    """

    def __init__(self, source_gadget, box, pos):
        """
        Initialize the command.

        @param source_gadget:
        @type source_gadget: L{gazpacho.gadget.Gadget}
        @param box: the box into which the widget should be inserted
        @type box: L{gazpacho.gadget.Gadget}
        @param pos: the position where the widget will be inserted
        @type pos: int
        """
        description = _("Create - extend")
        AppendGadgetCommand.__init__(self, source_gadget, box, pos, description)
        self._undo = False

    def execute(self):
        """
        Execute the command.
        """
        if self._undo:
            self.remove_execute()
        else:
            self.insert_execute()

        self._undo = not self._undo


class ExtendGadgetCommand(Command):
    """
    Command for replacing a target widget with a box containing a copy
    of both the target and the source.
    """

    def __init__(self, source_gadget, target_gadget, location,
                 description=None):
        """
        Initialize the command. Note that the source_gadget has to be
        a new widget that is not already in use in the project.

        @param source_gadget: the widget to add
        @type source_gadget: L{gazpacho.gadget.Gadget}
        @param target_gadget: the widget to replace
        @type target_gadget: L{gazpacho.gadget.Gadget}
        @param location: Where to put the source in relation to the target
        @type location: (constant value)
        """
        Command.__init__(self, description)

        self._project = target_gadget.project
        self._gtk_source = source_gadget.widget
        self._gtk_target = target_gadget.widget

        # Create the box with a copy of the target
        target_copy = target_gadget.copy(keep_name=True)

        self._gtk_box = self._create_box(source_gadget.widget,
                                         target_copy.widget, location)


    def _create_box(self, gtk_source, gtk_target, location):
        """
        Create a Box containing the widgets.

        @param gtk_source: the gtk  widget to add
        @type gtk_source: gtk.Gadget
        @param gtk_target: the gtk widget to replace
        @type gtk_target: gtk.Gadget
        @param location: Where to put the source in relation to the target
        @type location: (constant value)
        """
        from gazpacho.dndhandlers import DND_POS_TOP, DND_POS_BOTTOM, \
             DND_POS_LEFT

        # Create a Box with size 2
        if location in [DND_POS_TOP, DND_POS_BOTTOM]:
            box_type = 'GtkVBox'
        else:
            box_type = 'GtkHBox'
        adaptor = widget_registry.get_by_name(box_type)
        box_gadget = Gadget(adaptor, self._project)
        box_gadget.create_widget(interactive=False)
        box_gadget.get_prop('size').set(2)

        # Add the source and target widgets
        children = box_gadget.widget.get_children()
        if location in [DND_POS_TOP, DND_POS_LEFT]:
            source_placeholder, target_placeholder = children[0], children[1]
        else:
            source_placeholder, target_placeholder = children[1], children[0]

        Gadget.replace(source_placeholder, gtk_source, box_gadget)
        Gadget.replace(target_placeholder, gtk_target, box_gadget)

        return box_gadget.widget

    def _replace(self, target, source):
        """
        Replace the target widget with the source widget.

        @param target: the target gtk widget
        @type target: gtk.Gadget
        @param source: the source gtk widget
        @type source: gtk.Gadget
        """
        parent = util.get_parent(target)
        Gadget.replace(target, source, parent)

        # Remove old widget
        target.hide()
        self._project.remove_widget(target)

        # Add new widget
        self._project.add_widget(source)
        source.show_all()

    def add_box(self):
        """
        Replace the original target widget with the box.
        """
        self._replace(self._gtk_target, self._gtk_box)
        self._project.selection.set(self._gtk_source)

    def remove_box(self):
        """
        Replace the box with the original target widget.
        """
        self._replace(self._gtk_box, self._gtk_target)
        self._project.selection.set(self._gtk_target)

class DragExtendCommand(ExtendGadgetCommand):
    """
    Command for replacing a target widget with a box containing a copy
    of both the target and the source.
    """

    def __init__(self, source_gadget, target_gadget, location, keep_source):
        """
        Initialize the command.

        @param source_gadget: the widget to add
        @type source_gadget: L{gazpacho.gadget.Gadget}
        @param target_gadget: the widget to replace
        @type target_gadget: L{gazpacho.gadget.Gadget}
        @param location: Where to put the source in relation to the target
        @type location: (constant value)
        @param keep_source: if true we should not remove the source
        widget, if false we should.
        @type keep_source: bool
        """
        source_copy = source_gadget.copy(target_gadget.project,
                                         not keep_source)
        ExtendGadgetCommand.__init__(self, source_copy, target_gadget, location,
                                     _("Drag - extend"))

        self._undo = False
        self._remove_cmd = None

        # Reuse the Cut command for removing the widget
        if not keep_source:
            self._remove_cmd = CommandCutPaste(source_gadget,
                                               source_gadget.project,
                                               None,
                                               True)

    def execute(self):
        """
        Execute the command.
        """
        if self._undo:
            self.remove_box()

        if self._remove_cmd:
            self._remove_cmd.execute()

        if not self._undo:
            self.add_box()

        self._undo = not self._undo

class CreateExtendCommand(ExtendGadgetCommand):
    """
    Command for replacing a target widget with a box containing a copy
    of both the target and the source.
    """

    def __init__(self, source_gadget, target_gadget, location):
        """
        Initialize the command.

        @param source_gadget: the widget to add
        @type source_gadget: L{gazpacho.gadget.Gadget}
        @param target_gadget: the widget to replace
        @type target_gadget: L{gazpacho.gadget.Gadget}
        @param location: Where to put the source in relation to the target
        @type location: (constant value)
        @param keep_source: if true we should not remove the source
        widget, if false we should.
        @type keep_source: bool
        """
        ExtendGadgetCommand.__init__(self, source_gadget, target_gadget,
                                     location, _("Create - extend"))
        self._undo = False

    def execute(self):
        """
        Execute the command.
        """
        if self._undo:
            self.remove_box()

        if not self._undo:
            self.add_box()

        self._undo = not self._undo

class CommandBoxInsertPlaceholder(FlipFlopCommandMixin, Command):
    """
    Insert a placeholder before or after the specified position
    in the box.
    """

    def __init__(self, box, pos, after):

        FlipFlopCommandMixin.__init__(self, True)

        if after:
            description = _("Insert after")
            pos += 1
        else:
            description = _("Insert before")

        Command.__init__(self, description)

        self._box = box
        self._pos = pos
        self._placeholder = None

    def _insert_execute(self):
        # create a placeholder and insert it at self._pos
        self._placeholder = Placeholder()
        self._box.widget.add(self._placeholder)
        self._box.widget.reorder_child(self._placeholder, self._pos)
    _execute_state1 = _insert_execute

    def _delete_execute(self):
        self._placeholder.destroy()
        self._placeholder = None
    _execute_state2 = _delete_execute

class CommandBoxDeletePlaceholder(Command):
    def __init__(self, placeholder):
        Command.__init__(self, description = _("Delete placeholder"))

        parent = get_parent(placeholder)

        self._project = parent.project
        self._placeholder = placeholder
        self._gtk_parent = parent.widget
        self._create = False

        children = self._gtk_parent.get_children()
        self._pos = children.index(placeholder)

    def _create_execute(self):
        """
        Insert the placeholder at self._pos.
        """
        self._gtk_parent.add(self._placeholder)
        self._gtk_parent.reorder_child(self._placeholder, self._pos)

    def _delete_execute(self):
        """
        Remove the placeholder.
        """
        self._gtk_parent.remove(self._placeholder)
        self._project.selection.clear()

    def execute(self):
        if self._create:
            self._create_execute()
        else:
            self._delete_execute()

        self._create = not self._create




"""From http://www.davidcramer.net/code/62/set-cookies-without-a-response-in-django.html

Used with permission.  Thank you David Cramer.

To install, add the prehandler to the beginning of your middleware list, and the posthandler to the end.  Then, you can use request.COOKIES.get|set|delete anywhere in the code and they'll be automatically added to the response.
"""

from Cookie import SimpleCookie, Morsel
import copy
 
class CookiePreHandlerMiddleware(object):
    """
    This middleware modifies request.COOKIES and adds a set and delete method.
 
    `set` matches django.http.HttpResponse.set_cookie
    `delete` matches django.http.HttpResponse.delete_cookie
 
    This should be the first middleware you load.
    """
    def process_request(self, request):
        cookies = CookieHandler()
        for k, v in request.COOKIES.iteritems():
            cookies[k] = str(v)
        request.COOKIES = cookies
        request._orig_cookies = copy.deepcopy(request.COOKIES)
 
class CookiePostHandlerMiddleware(object):
    """
    This middleware modifies updates the response will all modified cookies.
 
    This should be the last middleware you load.
    """
    def process_response(self, request, response):
        if hasattr(request, '_orig_cookies') and request.COOKIES != request._orig_cookies:
            for k,v in request.COOKIES.iteritems():
                if request._orig_cookies.get(k) != v:
                    dict.__setitem__(response.cookies, k, v)
        return response
 
class StringMorsel(Morsel):
    def __str__(self):
        return self.value
 
    def __eq__(self, a):
        if isinstance(a, str):
            return str(self) == a
        elif isinstance(a, Morsel):
            return a.output() == self.output()
        return False
 
    def __ne__(self, a):
        if isinstance(a, str):
            return str(self) != a
        elif isinstance(a, Morsel):
            return a.output() != self.output()
        return True
 
    def __repr__(self):
        return str(self)
 
class CookieHandler(SimpleCookie):
    def __set(self, key, real_value, coded_value):
        """Private method for setting a cookie's value"""
        M = self.get(key, StringMorsel())
        M.set(key, real_value, coded_value)
        dict.__setitem__(self, key, M)
 
    def __setitem__(self, key, value):
        """Dictionary style assignment."""
        rval, cval = self.value_encode(value)
        self.__set(key, rval, cval)
 
    def set(self, key, value='', max_age=None, expires=None, path='/', domain=None, secure=None):
        self[key] = value
        for var in ('max_age', 'path', 'domain', 'secure', 'expires'):
            val = locals()[var]
            if val is not None:
                self[key][var.replace('_', '-')] = val
 
    def delete(self, key, path='/', domain=None):
        self[key] = ''
        if path is not None:
            self[key]['path'] = path
        if domain is not None:
            self[key]['domain'] = domain
        self[key]['expires'] = 0
        self[key]['max-age'] = 0


from django import http
from django.contrib.auth import REDIRECT_FIELD_NAME
from django.contrib.auth.decorators import login_required
from django.core import urlresolvers
from django.shortcuts import render_to_response
from django.template import RequestContext
from django.utils.translation import ugettext_lazy as _
from livesettings import config_value, config_get_group, SettingNotSet
from satchmo_store.contact import signals, CUSTOMER_ID
from satchmo_store.contact.forms import ExtendedContactInfoForm
from satchmo_store.contact.models import Contact
from satchmo_store.shop.models import Config
import logging

log = logging.getLogger('satchmo_store.contact.views')

def view(request):
    """View contact info."""
    try:
        user_data = Contact.objects.get(user=request.user.id)
    except Contact.DoesNotExist:
        user_data = None

    contact_dict = {
        'user_data': user_data, 
    }

    signals.satchmo_contact_view.send(user_data, contact=user_data, contact_dict=contact_dict)
            
    context = RequestContext(request, contact_dict)

    return render_to_response('contact/view_profile.html',
                              context_instance=context)

view = login_required(view)

def update(request):
    """Update contact info"""

    init_data = {}
    shop = Config.objects.get_current()

    try:
        contact = Contact.objects.from_request(request, create=False)
    except Contact.DoesNotExist:
        contact = None
    

    if request.method == "POST":
        new_data = request.POST.copy()            
        form = ExtendedContactInfoForm(data=new_data, shop=shop, contact=contact, shippable=True,
            initial=init_data)

        if form.is_valid():
            if contact is None and request.user:
                contact = Contact(user=request.user)
            custID = form.save(contact=contact)
            request.session[CUSTOMER_ID] = custID
            redirect_to = request.REQUEST.get(REDIRECT_FIELD_NAME, '')
            if not redirect_to or '//' in redirect_to or ' ' in redirect_to:
                redirect_to = urlresolvers.reverse('satchmo_account_info')

            return http.HttpResponseRedirect(redirect_to)
        else:
            signals.satchmo_contact_view.send(contact, contact=contact, contact_dict=init_data)

    else:
        if contact:
            #If a person has their contact info, make sure we populate it in the form
            for item in contact.__dict__.keys():
                init_data[item] = getattr(contact,item)
            if contact.shipping_address:
                for item in contact.shipping_address.__dict__.keys():
                    init_data["ship_"+item] = getattr(contact.shipping_address,item)
            if contact.billing_address:
                for item in contact.billing_address.__dict__.keys():
                    init_data[item] = getattr(contact.billing_address,item)
            if contact.primary_phone:
                init_data['phone'] = contact.primary_phone.phone
            if contact.organization:
                init_data['organization'] = contact.organization.name
   
            
        signals.satchmo_contact_view.send(contact, contact=contact, contact_dict=init_data)
        form = ExtendedContactInfoForm(shop=shop, contact=contact, shippable=True, initial=init_data)

    init_data['form'] = form
    if shop.in_country_only:
        init_data['country'] = shop.sales_country
    else:
        countries = shop.countries()
        if countries and countries.count() == 1:
            init_data['country'] = countries[0]
    
    
    init_data['next'] = request.REQUEST.get(REDIRECT_FIELD_NAME, '')
    context = RequestContext(request, init_data)

    return render_to_response('contact/update_form.html',
                              context_instance=context)

update = login_required(update)




# Copyright (C) 2004,2005 by SICEm S.L.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

import gtk
import gobject
from kiwi.component import get_utility
from kiwi.ui.dialogs import BaseDialog
from kiwi.utils import gsignal

from gazpacho.app.bars import bar_manager
from gazpacho.command import FlipFlopCommandMixin, Command
from gazpacho.commandmanager import command_manager
from gazpacho.gaction import GAction, GActionGroup
from gazpacho.i18n import _
from gazpacho.interfaces import IGazpachoApp
from gazpacho.stockicons import StockIconList
from gazpacho.signalhandlers import SignalHandlerStorage
from gazpacho.util import select_iter

(COL_OBJECT,) = list(range(1))

class GActionsView(gtk.ScrolledWindow):
    """A GActionsView is basically a TreeView where the actions
    and action groups are shown.

    The data it shows is always a two level hierarchy tree, where
    toplevel nodes are GActionGroups and leaves are GActions.
    """
    gsignal('selection-changed', object)

    def __init__(self, app):
        gtk.ScrolledWindow.__init__(self)
        self.set_shadow_type(gtk.SHADOW_IN)

        self._handlers = SignalHandlerStorage()
        self._app = app
        self.project = None

        self._model = gtk.TreeStore(object)
        self._treeview = gtk.TreeView(self._model)
        self._treeview.set_headers_visible(False)
        self._treeview.connect('row-activated', self._row_activated_cb)
        self._treeview.connect('button-press-event', self._button_press_cb)
        self._treeview.connect('key-press-event', self._key_press_cb)
        self._treeview.get_selection().connect('changed',
                                               self._selection_changed_cb)

        column = gtk.TreeViewColumn()
        renderer1 = gtk.CellRendererPixbuf()
        column.pack_start(renderer1, expand=False)
        column.set_cell_data_func(renderer1, self._draw_action, 0)

        renderer2 = gtk.CellRendererText()
        column.pack_start(renderer2, expand=True)
        column.set_cell_data_func(renderer2, self._draw_action, 1)

        self._treeview.append_column(column)

        self.add(self._treeview)
        self._treeview.show()

    def set_project(self, project):
        if self.project:
            self._handlers.disconnect_all()

        self.project = project
        self._model.clear()

        if not project:
            return

        self._handlers.connect(project, 'add-action', self._add_action_cb)
        self._handlers.connect(project, 'remove-action',
                               self._remove_action_cb)
        self._handlers.connect(project, 'action-name-changed',
                               self._action_name_changed_cb)

        self._populate_model()

    def add_action(self, gaction):
        """Create an action and add it to the selected action
        group. This method will not create the action directly but
        delegate to the command manager."""
        if isinstance(gaction, GActionGroup):
            # create an action with the selected action group
            # as the parent
            parent = gaction
        else:
            # create a brother action of the selected action
            parent = gaction.parent

        dialog = GActionDialog(
            get_utility(IGazpachoApp).get_window(), parent, None)
        if dialog.run() == gtk.RESPONSE_OK:
            values = dialog.get_values()
            new_gact = GAction(parent, values['name'], values['label'],
                               values['short_label'], values['is_important'],
                               values['tooltip'], values['stock_id'],
                               values['callback'], values['accelerator'])
            cmd = CommandAddRemoveAction(parent, new_gact, True)
            command_manager.execute(cmd, self.project)

        dialog.destroy()

    def add_action_group(self):
        """Create and add an action group. This method will not create
        the action group directly but delegate to the command manager."""
        # nothing is selected, we create an action group
        dialog = GActionGroupDialog(
            get_utility(IGazpachoApp).get_window(), None)
        if dialog.run() == gtk.RESPONSE_OK:
            name = dialog.get_action_group_name()
            gaction_group = GActionGroup(name)
            cmd = CommandAddRemoveActionGroup(gaction_group, self.project, True)
            command_manager.execute(cmd, self.project)
        dialog.destroy()

    def remove_action(self, gaction):
        """Remove the action group. This method will not remove the
        action group directly but delegate to the command manager."""
        if isinstance(gaction, GActionGroup):
            cmd = CommandAddRemoveActionGroup(gaction, self.project, False)
        else:
            cmd = CommandAddRemoveAction(gaction.parent, gaction, False)

        command_manager.execute(cmd, self.project)

    def edit_action(self, gaction):
        """Edit the action or action group. This method will not edit
        it directly but delegate to the command manager."""
        if isinstance(gaction, GAction):
            dialog = GActionDialog(
                get_utility(IGazpachoApp).get_window(),
                gaction.parent, gaction)
            if dialog.run() == gtk.RESPONSE_OK:
                new_values = dialog.get_values()
                cmd = CommandEditAction(gaction, new_values, self.project)
                command_manager.execute(cmd, self.project)

            dialog.destroy()
        else:
            dialog = GActionGroupDialog(get_utility(IGazpachoApp).get_window(),
                                        gaction)
            if dialog.run() == gtk.RESPONSE_OK:
                new_name = dialog.get_action_group_name()
                cmd = CommandEditActionGroup(gaction, new_name, self.project)
                command_manager.execute(cmd, self.project)

            dialog.destroy()

    def _populate_model(self):
        for gaction_group in self.project.uim.get_action_groups():
            parent_iter = self._model.append(None, (gaction_group,))
            for gaction in gaction_group.get_actions():
                self._model.append(parent_iter, (gaction,))

    def _selection_changed_cb(self, selection, data=None):
        """Callback for selection changes in the tree view."""
        model, it = selection.get_selected()
        item = None
        if it:
            item = model[it][COL_OBJECT]
        self.emit('selection-changed', item)

    def _add_action_cb(self, project, action):
        new_iter = None
        if isinstance(action, GActionGroup):
            new_iter = self._model.append(None, (action,))
        elif isinstance(action, GAction):
            parent = action.parent
            ag_iter = self._find_action_group(parent)
            if ag_iter is not None:
                new_iter = self._model.append(ag_iter, (action,))

        if new_iter:
            select_iter(self._treeview, new_iter)

    def _find_action_group(self, action_group):
        model = self._model
        for row in model:
            if row[COL_OBJECT] == action_group:
                return row.iter

    def _find_action(self, parent_iter, gaction):
        model = self._model
        for row in model[parent_iter].iterchildren():
            if row[COL_OBJECT] == gaction:
                return row.iter

    def _remove_action_cb(self, project, gaction):
        if isinstance(gaction, GActionGroup):
            action_group_iter = self._find_action_group(gaction)
            if action_group_iter is not None:
                del self._model[action_group_iter]
        elif isinstance(gaction, GAction):
            parent_iter = self._find_action_group(gaction.parent)
            if parent_iter is not None:
                child_iter = self._find_action(parent_iter, gaction)
                if child_iter is not None:
                    del self._model[child_iter]

    def _action_name_changed_cb(self, project, gaction):
        action_iter = None
        if isinstance(gaction, GAction):
            parent_iter = self._find_action_group(gaction.parent)
            action_iter = self._find_action(parent_iter, gaction)
        elif isinstance(gaction, GActionGroup):
            action_iter = self._find_action_group(gaction)

        if action_iter is not None:
            path = self._model[action_iter].path
            self._model.row_changed(path, action_iter)

    def get_selected_action(self):
        selection = self._treeview.get_selection()
        model, model_iter = selection.get_selected()
        if model_iter:
            return model[model_iter][COL_OBJECT]

    def get_selected_action_group(self):
        selection = self._treeview.get_selection()
        model, model_iter = selection.get_selected()
        if model_iter:
            gaction = model[model_iter][COL_OBJECT]
            if isinstance(gaction, GActionGroup):
                return gaction

    def _draw_action(self, column, cell, model, iter, model_column):
        action = model[iter][COL_OBJECT]
        if model_column == 0:
            prop = 'stock-id'
            if isinstance(action, GActionGroup):
                data = 'gtk-execute'
            else:
                data = action.stock_id or ''
        elif model_column == 1:
            prop = 'text'
            data = action.name

        cell.set_property(prop, data)

    def _row_activated_cb(self, treeview, path, column):
        gaction = self._model[path][COL_OBJECT]
        self.edit_action(gaction)

    # Keybord interaction callback
    def _key_press_cb(self, view, event):
        """Callback for handling key press events. Right now it's only
        used for deleting actions and action groups."""
        if event.keyval in [gtk.keysyms.Delete, gtk.keysyms.KP_Delete]:
            bar_manager.activate_action('Delete')

    def _button_press_cb(self, view, event):
        """Callback for handling mouse clicks. It is used to show a
        context menu."""
        if event.button != 3:
            return False

        # No need for a context menu if there is no project
        if not self.project:
            return False

        result = view.get_path_at_pos(int(event.x), int(event.y))
        path = None
        if result:
            path = result[0]
            view.set_cursor(path)
            view.grab_focus()

        popup = bar_manager.get_widget('/ActionPopup')
        popup.popup(None, None, None, event.button, event.time)
        return True

gobject.type_register(GActionsView)

class GActionDialog(BaseDialog):
    """This Dialog allows the creation and edition of a GAction."""
    def __init__(self, toplevel, action_group, edit_action=None):
        BaseDialog.__init__(self, parent=toplevel)

        # If we are currently editing the accelerator
        self._accel_editing = False

        # All actions that belong to the action group, except the
        # action we're editing
        self._available_actions = {}
        for action in action_group.get_actions():
            if action == edit_action:
                continue
            self._available_actions[action.name] = action

        if edit_action:
            title = _('Edit Action')
            icon = gtk.STOCK_OK
        else:
            title = _('Add Action')
            icon = gtk.STOCK_ADD
        self.set_title(title)
        self.add_button(gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL)
        self._ok_button = self.add_button(icon, gtk.RESPONSE_OK)

        self.set_border_width(6)
        self.set_resizable(False)
        self.set_has_separator(False)
        self.vbox.set_spacing(6)

        self._create_widgets()

        self.action = edit_action

        self.set_default_response(gtk.RESPONSE_OK)

        # Set mnemonic callback for the labels in the buttons.
        for widget in self.action_area.get_children():
            hbox = widget.child.child
            label = hbox.get_children()[1]
            keyval = label.get_property('mnemonic-keyval')
            label.connect('mnemonic-activate',
                          self._on_mnemonic_activate, keyval)

        self.vbox.show_all()

    def _create_widgets(self):

        size_group = gtk.SizeGroup(gtk.SIZE_GROUP_HORIZONTAL)

        # id
        hbox = gtk.HBox(spacing=6)
        label = self._create_label('_Id', size_group, hbox)
        self.id = gtk.Entry()
        self.id.set_activates_default(True)
        self.id.connect('changed', self._on_id_entry_changed)

        label.set_mnemonic_widget(self.id)
        hbox.pack_start(self.id)
        self.vbox.pack_start(hbox)

        # label
        hbox = gtk.HBox(spacing=6)
        label = self._create_label(_('_Label'), size_group, hbox)
        self.label = gtk.Entry()
        self.label.set_activates_default(True)
        label.set_mnemonic_widget(self.label)
        hbox.pack_start(self.label)
        self.vbox.pack_start(hbox)

        # short label
        hbox = gtk.HBox(spacing=6)
        label = self._create_label(_('S_hort Label'), size_group, hbox)
        self.short_label = gtk.Entry()
        self.short_label.set_activates_default(True)
        label.set_mnemonic_widget(self.short_label)
        hbox.pack_start(self.short_label)
        self.vbox.pack_start(hbox)

        # stock
        hbox = gtk.HBox(spacing=6)
        label = self._create_label(_('_Stock'), size_group, hbox)
        self.stock_enabled = gtk.CheckButton()
        self.stock_enabled.connect('toggled', self._on_stock_check_toggled)
        label.set_mnemonic_widget(self.stock_enabled)
        hbox.pack_start(self.stock_enabled, False, False)
        self.stock = StockIconList()
        self.stock.connect('changed', self._on_stock_changed)
        hbox.pack_start(self.stock)

        self.vbox.pack_start(hbox)

        # is_important
        hbox = gtk.HBox(spacing=6)
        label = self._create_label(_('I_mportant'), size_group, hbox)
        self.is_important = gtk.CheckButton()
        label.set_mnemonic_widget(self.is_important)
        hbox.pack_start(self.is_important, False, False)
        self.vbox.pack_start(hbox)

        # accelerator
        hbox = gtk.HBox(spacing=6)
        label = self._create_label(_('Accele_rator'), size_group, hbox)
        self.accelerator = gtk.Entry()
        self.accelerator.set_activates_default(True)
        self.accelerator.set_text('Press a key combination')
        self.accelerator.connect('key-press-event',
                                        self._on_accelerator_entry_key_press)
        self.accelerator.connect('button-press-event',
                                 self._on_accelerator_entry_button_press)
        self.accelerator.connect('focus-in-event',
                                 self.on_accelerator_focus_in)
        self.accelerator.connect('focus-out-event',
                                 self.on_accelerator_focus_out)
        label.set_mnemonic_widget(self.accelerator)
        hbox.pack_start(self.accelerator)

        self.vbox.pack_start(hbox)

        # tooltip
        hbox = gtk.HBox(spacing=6)
        label = self._create_label(_('_Tooltip'), size_group, hbox)
        self.tooltip = gtk.Entry()
        self.tooltip.set_activates_default(True)
        label.set_mnemonic_widget(self.tooltip)
        hbox.pack_start(self.tooltip)
        self.vbox.pack_start(hbox)

        # callback
        hbox = gtk.HBox(spacing=6)
        label = self._create_label(_('Call_back'), size_group, hbox)
        self.callback = gtk.Entry()
        self.callback.set_activates_default(True)
        label.set_mnemonic_widget(self.callback)
        hbox.pack_start(self.callback)
        self.vbox.pack_start(hbox)

    def _create_label(self, text, size_group, hbox):
        label = gtk.Label()
        label.set_markup_with_mnemonic(text+':')
        label.set_alignment(0.0, 0.5)

        keyval = label.get_property('mnemonic-keyval')
        label.connect('mnemonic-activate', self._on_mnemonic_activate, keyval)

        size_group.add_widget(label)
        hbox.pack_start(label, False, False)
        return label

    def on_accelerator_focus_in(self, widget, event):
        """
        Callback for when the accelerator entry gets focus.
        """
        self._accel_editing = True

    def on_accelerator_focus_out(self, widget, event):
        """
        Callback for when the accelerator entry looses focus.
        """
        self._accel_editing = False

    def _on_mnemonic_activate(self, widget, group_cycling, keyval):
        """
        Callback for the mnemonic activate signal. If we're currently
        editing the accelerator we have to disable the usual mnemonic
        action and use the keyval for as an accelerator value.
        """
        if self._accel_editing:
            key = gtk.accelerator_name(keyval, self.mnemonic_modifier)
            self.accelerator.set_text(key)
            return True

        return False

    def _on_accelerator_entry_key_press(self, entry, event):
        """Callback for handling key events in the accelerator
        entry. Accelerators are added by pressing Ctrl, Shift and/or
        Alt in combination with the desired key. Delete and Backspace
        can be used to clear the entry. No other types of editing is
        possible."""

        # Tab must be handled as normal. Otherwise we can't move from
        # the entry.
        if event.keyval == gtk.keysyms.Tab:
            return False

        modifiers = event.get_state() & gtk.accelerator_get_default_mod_mask()
        modifiers = int(modifiers)

        # Check if we should clear the entry
        clear_keys = [gtk.keysyms.Delete,
                      gtk.keysyms.KP_Delete,
                      gtk.keysyms.BackSpace]

        if modifiers == 0:
            if event.keyval in clear_keys:
                entry.set_text('')
            return True

        # Check if the accelerator is valid and add it to the entry
        if gtk.accelerator_valid(event.keyval, modifiers):
            accelerator = gtk.accelerator_name(event.keyval, modifiers)
            entry.set_text(accelerator)

        return True

    def _on_accelerator_entry_button_press(self, entry, event):
        """Callback for handling mouse clicks on the accelerator
        entry. It is used to show a context menu with a clear item."""
        if event.button != 3:
            return False

        clear_item = gtk.MenuItem(_('_Clear'), True)
        if entry.get_text():
            clear_item.connect_object('activate',
                                      lambda entry: entry.set_text(''),
                                      entry)
        else:
            clear_item.set_property('sensitive', False)

        menu = gtk.Menu()
        menu.append(clear_item)
        menu.show_all()
        menu.popup(None, None, None, event.button, event.time)
        return True

    def _on_id_entry_changed(self, entry):
        """
        Callback used for disabling the ok button if the name of the
        action is already in use for the specified action group.

        @param entry: the id entry
        @type entry: gtk.Entry
        """
        name = entry.get_text()
        self._ok_button.set_sensitive(name not in self._available_actions)

    def _on_stock_changed(self, stock, stock_id):
        self._update_button()

    def _update_button(self):
        stock_id = self.stock.get_stock_id()
        empty = self.stock_enabled.get_active() and not stock_id
        self._ok_button.set_sensitive(not empty)

    def _on_stock_check_toggled(self, button):
        active = button.get_active()
        self.stock.set_sensitive(active)
        self._update_button()

    def _clear_widgets(self):
        self.id.set_text('')
        self.label.set_text('')
        self.short_label.set_text('')
        self.is_important.set_active(False)
        self.stock_enabled.set_active(False)
        self.stock.reset()
        self.accelerator.set_text('')
        self.tooltip.set_text('')
        self.callback.set_text('')

    def _load_widgets(self):
        self.id.set_text(self._action.name)
        self.label.set_text(self._action.label)
        self.short_label.set_text(self._action.short_label or '')
        self.is_important.set_active(self._action.is_important == True)
        if self._action.stock_id is None:
            enabled = False
        else:
            enabled = True
            self.stock.set_stock_id(self._action.stock_id)
        self.stock_enabled.set_active(enabled)

        self.accelerator.set_text(self._action.accelerator or '')
        self.tooltip.set_text(self._action.tooltip or '')
        self.callback.set_text(self._action.callback or '')

    def _fill_action(self):
        self._action.name = self.id.get_text()
        self._action.label = self.label.get_text()
        self._action.short_label = self.short_label.get_text()
        self._action.is_important = self.is_important.get_active()
        if self.stock_enabled.get_active():
            self._action.stock_id = self.stock.get_stock_id()
        else:
            self._action.stock_id = None
        self._action.accelerator = self.accelerator.get_text()
        self._action.tooltip = self.tooltip.get_text()
        self._action.callback = self.callback.get_text()

    def get_action(self):
        self._fill_action()
        return self._action

    def set_action(self, gaction):
        if gaction is None:
            self._clear_widgets()
            self._action = GAction(None)
        else:
            self._action = gaction
            self._load_widgets()

    action = property(get_action, set_action)

    def get_values(self):
        values = dict(name=self.id.get_text(),
                      label=self.label.get_text(),
                      short_label=self.short_label.get_text(),
                      is_important=self.is_important.get_active(),
                      accelerator=self.accelerator.get_text(),
                      tooltip=self.tooltip.get_text(),
                      callback=self.callback.get_text())
        stock_id = None
        if self.stock_enabled.get_active():
            stock_id = self.stock.get_stock_id()
        values['stock_id'] = stock_id

        return values

gobject.type_register(GActionDialog)

class GActionGroupDialog(BaseDialog):
    """This Dialog allows the creation and edition of a GActionGroup."""
    def __init__(self, toplevel=None, action_group=None):
        BaseDialog.__init__(self, parent=toplevel)

        if action_group is None:
            title = _('Add Action Group')
            stock_ok = gtk.STOCK_ADD
        else:
            title = _('Edit Action Group')
            stock_ok = gtk.STOCK_OK
        self.add_button(gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL)
        self.add_button(stock_ok, gtk.RESPONSE_OK)
        self.set_default_response(gtk.RESPONSE_OK)
        self.set_title(title)
        self.set_border_width(6)
        self.set_resizable(False)
        self.set_has_separator(False)
        self.vbox.set_spacing(6)

        # id
        hbox = gtk.HBox(spacing=6)
        label = gtk.Label()
        label.set_markup_with_mnemonic(_('_Id')+':')
        label.set_alignment(0.0, 0.5)
        hbox.pack_start(label, False, False)
        self._id = gtk.Entry()
        self._id.set_activates_default(True)
        label.set_mnemonic_widget(self._id)
        hbox.pack_start(self._id)
        self.vbox.pack_start(hbox)

        self.action_group = action_group

        self.vbox.show_all()

    def _clear_widgets(self):
        self._id.set_text('')

    def _load_widgets(self):
        self._id.set_text(self._action_group.name)

    def _fill_action_group(self):
        self._action_group.name = self._id.get_text()

    def get_action_group(self):
        self._fill_action_group()
        return self._action_group

    def set_action_group(self, action_group):
        if action_group is None:
            self._clear_widgets()
            self._action_group = GActionGroup(None)
        else:
            self._action_group = action_group
            self._load_widgets()

    action_group = property(get_action_group, set_action_group)

    def get_action_group_name(self):
        return self._id.get_text()

gobject.type_register(GActionGroupDialog)

class CommandAddRemoveAction(FlipFlopCommandMixin, Command):
    def __init__(self, parent, gact, add):
        FlipFlopCommandMixin.__init__(self, add)
        if add:
            description = _('Add action %s') % gact.name
        else:
            description = _('Remove action %s') % gact.name

        Command.__init__(self, description)

        self.gact = gact
        self.parent = parent

    def _add_execute(self):
        self.parent.add_action(self.gact)
    _execute_state1 = _add_execute

    def _remove_execute(self):
        self.parent.remove_action(self.gact)
    _execute_state2 = _remove_execute

class CommandEditAction(Command):
    def __init__(self, gact, new_values, project):
        Command.__init__(self, _('Edit action %s') % gact.name)

        self.new_values = new_values
        self.gact = gact
        self.project = project

    def execute(self):
        old_values = {
            'name' : self.gact.name,
            'label': self.gact.label,
            'short_label': self.gact.short_label,
            'is_important': self.gact.is_important,
            'stock_id': self.gact.stock_id,
            'tooltip': self.gact.tooltip,
            'accelerator': self.gact.accelerator,
            'callback': self.gact.callback,
            }
        self.gact.name = self.new_values['name']
        self.gact.label = self.new_values['label']
        self.gact.short_label = self.new_values['short_label']
        self.gact.is_important = self.new_values['is_important']
        self.gact.stock_id = self.new_values['stock_id']
        self.gact.tooltip = self.new_values['tooltip']
        self.gact.accelerator = self.new_values['accelerator']
        self.gact.callback = self.new_values['callback']
        self.gact.parent.update_action(self.gact, old_values['name'])
        self.new_values = old_values
        self.project.change_action_name(self.gact)

class CommandAddRemoveActionGroup(FlipFlopCommandMixin, Command):
    def __init__(self, gaction_group, project, add):
        FlipFlopCommandMixin.__init__(self, add)
        if add:
            description = _('Add action group %s') % gaction_group.name
        else:
            description = _('Remove action group %s') % gaction_group.name
        Command.__init__(self, description)

        self.project = project
        self.gaction_group = gaction_group
        self.gactions = self.gaction_group.get_actions()

    def _add_execute(self):
        self.project.add_action_group(self.gaction_group)
        for gaction in self.gactions:
            self.gaction_group.add_action(gaction)
        return self.gaction_group
    _execute_state1 = _add_execute

    def _remove_execute(self):
        self.project.remove_action_group(self.gaction_group)
        return self.gaction_group
    _execute_state2 = _remove_execute

class CommandEditActionGroup(Command):
    def __init__(self, gaction_group, new_name, project):
        description = _('Edit action group %s') % gaction_group.name
        Command.__init__(self, description)

        self.new_name = new_name
        self.gaction_group = gaction_group
        self.project = project

    def execute(self):
        old_name = self.gaction_group.name
        self.gaction_group.name = self.new_name
        self.new_name = old_name
        self.project.change_action_name(self.gaction_group)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Demo wiki using MoinMoinController request handler.

Copyright (C) 2010-2011 Etienne Robillard <erob@gthcfoundation.org>
All rights reserved.
"""

import sys, os
import logging

#from MoinMoin.server.server_wsgi import WsgiConfig
#XXX this require MoinMoin-1.8.9. The request package has been
#deprecated in MoinMoin-1.9.x.
from MoinMoin.request import request_wsgi

examples_dir = os.getcwd()
sys.path.extend([
    os.path.join(examples_dir, 'lib'),
    os.path.join(examples_dir, 'lib/site-packages')
    ])

import wikiapp
from notmm.http.httpserver import daemonize, get_bind_addr

# Add the wikiconfig module in sys.path if not present
if not 'wikiconfig' in sys.modules:
    sys.path.insert(-1, wikiapp.default_instance)
    import wikiconfig
    print "found correct wikiconfig!"

app_conf = wikiapp.global_conf['wikiapp']
#print app_conf

# Check if moin.url_prefix_static has been defined
# in development.ini, otherwise use the default (/moin_static184).
url_prefix_static = app_conf.get('moin.url_prefix_static', '/moin_static184')
wsgi_request_class = getattr(request_wsgi, 'Request')  
#response_class = HTTPResponse

from MoinMoin.server import Config as WsgiConfig

class WikiConfig(WsgiConfig):
    url_prefix_static = url_prefix_static

def main():
    #init the config instance
    config = WikiConfig()
    #init the main response handler
    wsgi_app = wikiapp.MoinMoinController(settings=config, app_conf={
        'django.settings_autoload' : False,
        'wsgi.request_class' : wsgi_request_class,
        'logging.disabled'   : True})
    
    # get the default inet address to bind the wsgi server to
    bind_addr = get_bind_addr(wikiapp.global_conf)
    
    # start the wsgi server
    s = daemonize(wsgi_app, bind_addr)
    s.serve_forever()

if __name__ == '__main__':
    main()



import random
from livesettings import config_value

def generate_certificate_code():
    """Creates a code, formatted according to the shop owner's preference as set in the config system."""
    charset = config_value('PAYMENT_GIFTCERTIFICATE', 'CHARSET')
    format = config_value('PAYMENT_GIFTCERTIFICATE', 'FORMAT')
    return generate_code(charset, format)

def generate_code(charset, format):
    """Creates the actual code.  Split out for ease of testing."""
    
    out = []
    for c in format.strip():
        if c=="^":
            out.extend(random.sample(charset, 1))
        else:
            out.append(c)
    
    return "".join(out)


import gtk
import gobject
from kiwi.component import get_utility
from kiwi.environ import environ
from kiwi.ui.dialogs import BaseDialog, warning
from kiwi.utils import gsignal

from gazpacho.app.bars import bar_manager
from gazpacho.command import FlipFlopCommandMixin, Command
from gazpacho.commandmanager import command_manager
from gazpacho.i18n import _
from gazpacho.interfaces import IGazpachoApp
from gazpacho.sizegroup import GSizeGroup
from gazpacho.util import select_iter
from gazpacho.gadget import Gadget
from gazpacho.signalhandlers import SignalHandlerStorage
from gazpacho.sizegroup import safe_to_add_gadgets

EDITOR_OBJECT_COLUMN, EDITOR_CALLBACK_IDS_COLUMN = list(range(2))

class SizeGroupView(gtk.ScrolledWindow):
    """
    The sizegroup view/editor that displays the sizegroup tree.
    """

    gsignal('selection-changed', object)

    def __init__(self, app):
        """
        Initialize the size group editor.

        @param app: the application
        @type app: gazpacho.application.Application
        """
        gtk.ScrolledWindow.__init__(self)
        self.set_shadow_type(gtk.SHADOW_IN)

        self._app = app
        self.project = None
        self._handlers = SignalHandlerStorage()
        self._mode_icons = self._load_mode_icons()

        self._treeview = self._create_treeview()
        self._model = gtk.TreeStore(object, object)
        self._treeview.set_model(self._model)

        # Connect the callbacks
        self._treeview.connect('button-press-event',
                               self._on_treeview_button_press_event)
        self._treeview.connect('key-press-event',
                               self._on_treeview_key_press_event)
        selection = self._treeview.get_selection()
        selection.connect('changed', self._on_treeview_selection_changed)

        self.add(self._treeview)
        self._treeview.show()

    #
    # Private methods
    #

    def _create_treeview(self):
        """
        Create the TreeView.
        """
        treeview = gtk.TreeView()
        treeview.set_headers_visible(False)

        column = gtk.TreeViewColumn()

        renderer1 = gtk.CellRendererPixbuf()
        column.pack_start(renderer1, False)
        column.set_cell_data_func(renderer1, self._draw_cell_icon)

        renderer2 = gtk.CellRendererText()
        column.pack_start(renderer2)
        column.set_cell_data_func(renderer2, self._draw_cell_data)

        treeview.append_column(column)
        return treeview

    def _load_mode_icons(self):
        """
        Load the pixbufs that are to be used as sizegroup mode icons.

        Mode constants:
         - gtk.SIZE_GROUP_HORIZONTAL
         - gtk.SIZE_GROUP_VERTICAL
         - gtk.SIZE_GROUP_BOTH

        @return: a mapping of sizegroup mode constants to pixbufs
        @rtype: map
        """
        icon_names = {gtk.SIZE_GROUP_HORIZONTAL: 'sizegroup-horizontal.png',
                      gtk.SIZE_GROUP_VERTICAL: 'sizegroup-vertical.png',
                      gtk.SIZE_GROUP_BOTH: 'sizegroup-both.png'}
        icon_map = {}
        for key, name in icon_names.items():
            pixbuf = None
            filename = environ.find_resource('pixmap', name)
            if filename:
                pixbuf = gtk.gdk.pixbuf_new_from_file(filename)
            icon_map[key] = pixbuf

        return icon_map

    def _draw_cell_icon(self, tree_column, cell, model, model_iter):
        """
        Draw the sizegroup or gadget icon.
        """
        row = model[model_iter]
        item = row[EDITOR_OBJECT_COLUMN]

        if isinstance(item, GSizeGroup):
            pixbuf = self._mode_icons.get(item.mode)
        else:
            pixbuf = item.adaptor.pixbuf

        cell.set_property('pixbuf', pixbuf)


    def _draw_cell_data(self, column, cell, model, model_iter):
        """
        Display the name of the sizegroup or widget
        """
        sizegroup = model[model_iter][EDITOR_OBJECT_COLUMN]
        text = sizegroup.name
        cell.set_property('text', text)


    def _populate_model(self):
        """
        Fill the view with data from the current project. This will
        add all sizegroups that belong to the project and also connect
        the necessary callbacks.

        Note, it is important that the project has been set befor
        calling this method.
        """
        sizegroup_iter = None
        for sizegroup in self.project.sizegroups:
            sizegroup_iter = self._add_sizegroup(sizegroup)

        if sizegroup_iter:
            select_iter(self._treeview, sizegroup_iter)

    def _add_sizegroup(self, sizegroup):
        """
        Add a sizegroup and its widgets to the treeview. This will the
        necessary callbacks as well.

        @param sizegroup: the sizegroup that should be added
        @type sizegroup: gazpacho.sizegroup.GSizeGroup

        @return: the iter pointing to the sizegroup that was added
        @rtype: gtk.TreeIter
        """
        id1 = sizegroup.connect('gadgets-added',
                                self._on_sizegroup_gadgets_added)
        id2 = sizegroup.connect('gadgets-removed',
                                self._on_sizegroup_gadgets_removed)
        id3 = sizegroup.connect('name-changed',
                                self._on_sizegroup_name_changed)

        sizegroup_iter = self._model.append(None, (sizegroup, (id1, id2, id3)))
        self._add_gadgets(sizegroup_iter, sizegroup.get_gadgets())
        return sizegroup_iter

    def _add_gadgets(self, sizegroup_iter, gadgets):
        """
        Add a number of gadgets to the treeview. This will the
        necessary callbacks as well.

        @param sizegroup_iter: the parent iter where the gadgets
          should be added
        @type sizegroup_iter: gtk.TreeIter
        @param gadgets: the gadgets that should be added
        @type gadgets: list (of gazpacho.gadget.Gadget)

        @return: the iter pointing to the last gadget that was added
        @rtype: gtk.TreeIter
        """
        sizegroup = self._model[sizegroup_iter][EDITOR_OBJECT_COLUMN]

        gadget_iter = None
        for gadget in gadgets:
            # FIXME: Disconnect callback
            gadget.widget.connect('notify::name',
                                  self._on_widget_notify_name,
                                  gadget,
                                  sizegroup)
            row = (gadget, ())
            gadget_iter = self._model.append(sizegroup_iter, row)

        return gadget_iter

    def _disconnect_item_callbacks(self, model_iter):
        """
        Disconnect all callbacks for the item (sizegroup or gadget)
        that the iter is pointing to.

        @param model_iter: the iter pointing to the item who's
                           callbacks should be disconnected
        @type model_iter: gtk.TreeIter
        """
        row = self._model[model_iter]
        item = row[EDITOR_OBJECT_COLUMN]
        ids = row[EDITOR_CALLBACK_IDS_COLUMN]
        for callback_id in ids:
            item.disconnect(callback_id)

    def _disconnect_all_callbacks(self, sizegroup_iter):
        """
        Disconnect all callbacks for a sizegroup and all its gadgets.

        @param sizegroup_iter: the iter pointing to the sizegroup
        who's callbacks should be disconnected
        @type sizegroup_iter: gtk.TreeIter
        """
        gadget_iter = self._model.iter_children(sizegroup_iter)
        while gadget_iter:
            self._disconnect_item_callbacks(gadget_iter)
            gadget_iter = self._model.iter_next(gadget_iter)

        self._disconnect_item_callbacks(sizegroup_iter)

    def _find_sizegroup(self, sizegroup):
        """
        Find a certain sizegroup in the TreeStore.

        @param sizegroup: the sizegroup that the gadget belong to
        @type sizegroup: gazpacho.sizegroup.GSizeGroup

        @return: the iter pointing to the sizegroup or None if it wasn't found
        @rtype: gtk.TreeIter
        """
        model = self._model
        for sizegroup_row in model:
            if sizegroup_row[EDITOR_OBJECT_COLUMN] == sizegroup:
                return sizegroup_row.iter
        return None

    def _find_gadget(self, sizegroup, gadget):
        """
        Find a certain gadget in the TreeStore.

        @param sizegroup: the sizegroup that the gadget belong to
        @type sizegroup: gazpacho.sizegroup.GSizeGroup
        @param gadget: the gadget that we're looking for
        @type gadget: gazpacho.gadget.Gadget

        @return: the iter pointing to the gadget or None if it wasn't found
        @rtype: gtk.TreeIter
        """
        sizegroup_iter = self._find_sizegroup(sizegroup)
        if not sizegroup_iter:
            return None

        model = self._model
        for gadget_row in model[sizegroup_iter].iterchildren():
            if gadget_row[EDITOR_OBJECT_COLUMN] == gadget:
                return gadget_row.iter
        return None

    #
    # Editor callbacks
    #

    def _on_treeview_selection_changed(self, selection):
        """
        Callback for selection changes in the tree view.

        @param selection: the selection that has changed
        @type selection: gtk.TreeSelection
        """
        model, model_iter = selection.get_selected()
        item = None
        if model_iter:
            item = model[model_iter][EDITOR_OBJECT_COLUMN]
        self.emit('selection-changed', item)

    def _on_treeview_key_press_event(self, view, event):
        """
        Callback for handling key press events. Right now it's only
        used for deleting sizegroups and gadgets.

        @param view: the sizegroup treeview
        @type view: gtk.TreeView
        @param event: the event that was triggered
        @type event: gtk.gdk.Event
        """
        if event.keyval in [gtk.keysyms.Delete, gtk.keysyms.KP_Delete]:
            bar_manager.activate_action('Delete')

    def _on_treeview_button_press_event(self, view, event):
        """
        Callback for handling mouse clicks. It is used to show a
        context menu.

        @param view: the sizegroup treeview
        @type view: gtk.TreeView
        @param event: the event that was triggered
        @type event: gtk.gdk.Event
        """
        if event.button != 3:
            return False

        # No need for a context menu if there is no project
        if not self.project:
            return False

        result = view.get_path_at_pos(int(event.x), int(event.y))
        if not result:
            return
        path = result[0]

        # Select the row
        view.set_cursor(path)
        view.grab_focus()

        # Show popup
        popup = bar_manager.get_widget('/SizeGroupPopup')
        popup.popup(None, None, None, event.button, event.time)
        return True

    #
    # General callbacks
    #

    def _on_project_add_sizegroup(self, project, sizegroup):
        """
        Callback that is executed when a sizegroup is added to a
        project.

        @param project: the project that the sizegroup was added to
        @type project: gazpacho.project.Project
        @param sizegroup: the sizegroup that have been added
        @type sizegroup: gazpacho.sizegroup.GSizeGroup
        """
        sizegroup_iter = self._add_sizegroup(sizegroup)
        if sizegroup_iter:
            select_iter(self._treeview, sizegroup_iter)

    def _on_project_remove_sizegroup(self, project, sizegroup):
        """
        Callback that is executed when a sizegroup is removed from a
        project.

        @param project: the project that the sizegroup was removed from
        @type project: gazpacho.project.Project
        @param sizegroup: the sizegroup that have been removed
        @type sizegroup: gazpacho.sizegroup.GSizeGroup
        """
        sizegroup_iter = self._find_sizegroup(sizegroup)
        if not sizegroup_iter:
            return

        self._disconnect_all_callbacks(sizegroup_iter)
        del self._model[sizegroup_iter]

    def _on_sizegroup_name_changed(self, sizegroup):
        """
        Callback that is executed when the name of a sizegroup has
        changed. This will make sure that the editor is updated.

        @param sizegroup: the sizegroup who's name has changed
        @type sizegroup: gazpacho.sizegroup.GSizeGroup
        """
        sizegroup_iter = self._find_sizegroup(sizegroup)
        if sizegroup_iter:
            path = self._model[sizegroup_iter].path
            self._model.row_changed(path, sizegroup_iter)


    def _on_sizegroup_gadgets_added(self, sizegroup, gadgets):
        """
        Callback that is executed when gadgets are added to a
        sizegroup. This will add the gadgets to the editor.

        @param sizegroup: the sizegroup that the gadgets belong to
        @type sizegroup: gazpacho.sizegroup.GSizeGroup
        @param gadgets: the gadgets that have been added
        @type gadgets: list (of gazpacho.gadget.Gadget)
        """
        sizegroup_iter = self._find_sizegroup(sizegroup)
        if not sizegroup_iter:
            return

        gadget_iter = self._add_gadgets(sizegroup_iter, gadgets)
        if gadget_iter:
            select_iter(self._treeview, gadget_iter)

    def _on_sizegroup_gadgets_removed(self, sizegroup, gadgets):
        """
        Callback that is executed when gadgets are removed from a
        sizegroup. This will remove the gadgets from the editor.

        @param sizegroup: the sizegroup that the gadgets belong to
        @type sizegroup: gazpacho.sizegroup.GSizeGroup
        @param gadgets: the gadgets that have been removed
        @type gadgets: list (of gazpacho.gadget.Gadget)
        """
        for gadget in gadgets:
            gadget_iter = self._find_gadget(sizegroup, gadget)
            if gadget_iter:
                self._disconnect_item_callbacks(gadget_iter)
                del self._model[gadget_iter]

    def _on_widget_notify_name(self, widget, pspec, gadget, sizegroup):
        """
        Callback that is executed when the name property of a gadget
        has changed. This will make sure that the editor is updated.

        @param widget: the widget who's name has changed
        @type widget: gtk.Widget
        @param pspec: the gobject.GParamSpec of the property that was changed
        @type pspec: gobject.GParamSpec
        @param sizegroup: the sizegroup that the gadgets belong to
        @type sizegroup: gazpacho.sizegroup.GSizeGroup
        """
        gadget_iter = self._find_gadget(sizegroup, gadget)
        if gadget_iter:
            path = self._model[gadget_iter].path
            self._model.row_changed(path, gadget_iter)

    #
    # Public methods
    #
    def has_selected(self):
        """
        Check if a sizegroup or gadget is selected in the tree.
        """
        model_iter = self._treeview.get_selection().get_selected()[1]
        return model_iter is not None

    def set_project(self, project):
        """
        Set the current project.

        @param project: the current project
        @type project: gazpacho.project.project
        """
        if self.project:
            for sizegroup_row in self._model:
                self._disconnect_all_callbacks(sizegroup_row.iter)
            self._handlers.disconnect_all()
            self._model.clear()

        self.project = project

        if not project:
            return

        self._populate_model()
        self._handlers.connect(project, 'add-sizegroup',
                               self._on_project_add_sizegroup)
        self._handlers.connect(project, 'remove-sizegroup',
                               self._on_project_remove_sizegroup)

    def remove_selected_item(self):
        """
        Convenient method for removing the selecte sizegroup or gadget

        This method will not remove the action group directly but
        delegate to the command manager.
        """
        model, model_iter = self._treeview.get_selection().get_selected()
        if not model_iter:
            return

        row = model[model_iter]
        if row.parent:
            sizegroup = row.parent[EDITOR_OBJECT_COLUMN]
            gadgets = [row[EDITOR_OBJECT_COLUMN]]
        else:
            sizegroup = row[EDITOR_OBJECT_COLUMN]
            gadgets = sizegroup.get_gadgets()

        cmd = CommandAddRemoveSizeGroupGadgets(sizegroup, gadgets, self.project,
                                               False)
        command_manager.execute(cmd, self.project)

gobject.type_register(SizeGroupView)

def add_sizegroup_gadgets(project, sizegroup=None, gadgets=[]):
    """
    Convenient method for adding gadgets to a sizegroup.

    If no sizegroup is specified, a dialog will prompt the user to choose one.

    If no gadgets are specified, the current project selected widgets will be
    added.

    @param project: the current project
    @type project: gazpacho.project.Project
    """
    if not project.selection:
        return

    if sizegroup is None:
        window = get_utility(IGazpachoApp).get_window()
        dialog = SizeGroupDialog(window, project.sizegroups)
        if dialog.run() != gtk.RESPONSE_OK:
            dialog.destroy()
            return

        sizegroup = dialog.get_selected_sizegroup()
        dialog.destroy()

    if not gadgets:
        gadgets = [Gadget.from_widget(w) for w in project.selection]

    add_gadgets = []
    for gadget in gadgets:
        # We don't add a gadget that's already in the sizegroup
        if not sizegroup.has_gadget(gadget):
            add_gadgets.append(gadget)

    if gadgets and not safe_to_add_gadgets(sizegroup, add_gadgets):
        warning(_("Cannot add the widget"),
                _("It's not possible to add a gadget who has an ancestor"
                  " or child in the sizegroup."))
        return

    cmd = CommandAddRemoveSizeGroupGadgets(sizegroup, add_gadgets, project,
                                           True)
    command_manager.execute(cmd, project)

DIALOG_MODE_NAME_COLUMN, DIALOG_MODE_VALUE_COLUMN = list(range(2))

DIALOG_SIZEGROUP_NAME_COLUMN, DIALOG_SIZEGROUP_COLUMN = list(range(2))

class SizeGroupDialog(BaseDialog):
    """
    A dialog for choosing which sizegroup the widgets should be added
    to. It also offers the possibility to create a new sizegroup.
    """

    def __init__(self, parent, sizegroups):
        """
        Initialize the dialog.

        @param sizegroups: all available sizegroups
        @type sizegroups: list (of gazpacho.sizegroup.GSizeGroup)
        """
        BaseDialog.__init__(self, title=_('Add Size Group Widgets'),
                            parent=parent)

        # Dict of available sizegroups
        self._sizegroups = self._get_sizegroup_dict(sizegroups)

        # Existing sizegroups combo
        self._sizegroup_combo = None

        # New sizegroup name
        self._name_entry = None

        # New sizegroup modes combo
        self._mode_combo = None

        # Radio buttons
        self._selection_radio = None
        self._new_radio = None

        self.add_button(gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL)
        self._add_button = self.add_button(gtk.STOCK_ADD, gtk.RESPONSE_OK)

        self._create_content()

        # Set default name
        name = self._suggest_sizegroup_name()
        self._name_entry.set_text(name)

        self._name_entry.connect('changed', self._on_name_entry_changed)

        self.set_default_response(gtk.RESPONSE_OK)
        self.vbox.show_all()

    def _get_sizegroup_dict(self, sizegroups):
        """
        Get a dictionary that maps sizegroup names to sizegroups.

        @return: dict mapping name to sizegroup
        @rtype: dict (str: gazpacho.sizegroup.GSizeGroup)
        """
        group_dict = {}
        for sizegroup in sizegroups:
            group_dict[sizegroup.name] = sizegroup
        return group_dict

    #
    # Public methods
    #

    def get_selected_sizegroup(self):
        """
        Get the selected sizegroup. This might return an existing
        sizegroup or a new one.

        @return: the sizegroup
        @rtype: gazpacho.sizegroup.GSizeGroup
        """
        if self._selection_radio.get_active():
            model = self._sizegroup_combo.get_model()
            active = self._sizegroup_combo.get_active()
            sizegroup = model[active][DIALOG_SIZEGROUP_COLUMN]
        else:
            model = self._mode_combo.get_model()
            active = self._mode_combo.get_active()
            mode = model[active][DIALOG_MODE_VALUE_COLUMN]
            name =  self._name_entry.get_text()
            sizegroup = GSizeGroup(name, gtk.SizeGroup(mode))

        return sizegroup

    #
    # Private methods
    #

    def _create_content(self):
        """
        Create the dialog content.
        """
        size_group = gtk.SizeGroup(gtk.SIZE_GROUP_HORIZONTAL)

        # Existing sizegroup area
        self._selection_radio = gtk.RadioButton(None,
                                                _('Existing size group:'))
        self.vbox.pack_start(self._selection_radio, False, False)

        selection_area = self._create_sizegroup_selection_area(size_group)
        self.vbox.pack_start(selection_area, False, False)

        # New sizegroup area
        self._new_radio = gtk.RadioButton(self._selection_radio,
                                          _('New size group:'))
        self.vbox.pack_start(self._new_radio, False, False)

        new_sizegroup_area = self._create_new_sizegroup_area(size_group)
        self.vbox.pack_start(new_sizegroup_area, False, False)

        # Select which should be active
        if self._sizegroups:
            new_sizegroup_area.set_sensitive(False)
        else:
            self._selection_radio.set_sensitive(False)
            self._new_radio.set_active(True)
            selection_area.set_sensitive(False)

        # Add some handlers
        self._selection_radio.connect('toggled',
                                      self._on_radio_toggled, selection_area)
        self._new_radio.connect('toggled',
                                self._on_radio_toggled, new_sizegroup_area)


    def _create_sizegroup_selection_area(self, size_group):
        """
        Create the GUI for selecting an existing sizegroup.

        @param size_group: a SizeGroup for the labels
        @type size_group: gtk.SizeGroup
        """
        alignment = gtk.Alignment(1.0, 0.0, 1.0, 1.0)
        alignment.set_padding(0, 12, 48, 0)

        hbox = gtk.HBox()
        hbox.set_spacing(6)

        label = gtk.Label(('Size group:'))
        label.set_alignment(0.0, 0.5)
        size_group.add_widget(label)
        hbox.pack_start(label, False, False)

        liststore = gtk.ListStore(str, object)
        sizegroups = list(self._sizegroups.values())
        sizegroups.sort(lambda x, y: cmp(x.name, y.name))
        for group in sizegroups:
            liststore.append((group.name, group))
        self._sizegroup_combo = gtk.ComboBox(liststore)
        renderer = gtk.CellRendererText()
        self._sizegroup_combo.pack_start(renderer)
        self._sizegroup_combo.add_attribute(renderer, 'text', 0)

        self._sizegroup_combo.set_active(0)
        hbox.pack_start(self._sizegroup_combo, True, True)

        alignment.add(hbox)
        return alignment

    def _create_name_entry(self, size_group):
        """
        Create the sizegroup name entry area.

        @param size_group: a SizeGroup for the labels
        @type size_group: gtk.SizeGroup
        """
        hbox = gtk.HBox()
        hbox.set_spacing(6)

        label = gtk.Label(('Name:'))
        label.set_alignment(0.0, 0.5)
        size_group.add_widget(label)

        hbox.pack_start(label, False, False)

        self._name_entry = gtk.Entry()
        hbox.pack_start(self._name_entry, True, True)

        return hbox

    def _create_mode_entry(self, size_group):
        """
        Create the sizegroup name mode area.

        @param size_group: a SizeGroup for the labels
        @type size_group: gtk.SizeGroup
        """
        hbox = gtk.HBox()
        hbox.set_spacing(6)

        label = gtk.Label(('Mode:'))
        label.set_alignment(0.0, 0.5)
        size_group.add_widget(label)
        hbox.pack_start(label, False, False)

        liststore = gtk.ListStore(str, int)
        liststore.append((_('Horizontal'), gtk.SIZE_GROUP_HORIZONTAL))
        liststore.append((_('Vertical'), gtk.SIZE_GROUP_VERTICAL))
        liststore.append((_('Both'), gtk.SIZE_GROUP_BOTH))
        self._mode_combo = gtk.ComboBox(liststore)
        renderer = gtk.CellRendererText()
        self._mode_combo.pack_start(renderer)
        self._mode_combo.add_attribute(renderer, 'text', 0)

        self._mode_combo.set_active(0)
        hbox.pack_start(self._mode_combo, True, True)

        return hbox

    def _create_new_sizegroup_area(self, size_group):
        """
        Create the GUI for creating a new sizegroup.

        @param size_group: a SizeGroup for the labels
        @type size_group: gtk.SizeGroup
        """
        vbox = gtk.VBox()
        vbox.set_spacing(6)
        vbox.pack_start(self._create_name_entry(size_group), False, False)
        vbox.pack_start(self._create_mode_entry(size_group), False, False)

        alignment = gtk.Alignment(1.0, 0.0, 1.0, 1.0)
        alignment.set_padding(0, 12, 48, 0)
        alignment.add(vbox)

        return alignment

    def _suggest_sizegroup_name(self):
        """
        Get a unique name suggestion for the sizegroup. It will be
        something like 'sizegroup3'.

        @return: a unique name
        @rtype: str
        """
        default_name = "sizegroup%d"
        i = 1
        while True:
            name = default_name % i
            if name not in self._sizegroups:
                return name
            i += 1

    def _on_radio_toggled(self, button, widget):
        """
        Callback that is executed when the radio buttons are
        toggled. This will enable or disable parts of the gui.

        @param button: the radio button that was toggled
        @type button: gtk.RadioButton
        @param widget: the gui element that should be enabled or disabled
        @type widget: gtk.Widget
        """
        value = button.get_active()
        widget.set_sensitive(value)

    def _on_name_entry_changed(self, entry):
        """
        Callback that is executed when the text in the name entry is
        modified.

        @param entry: the name entry
        @type entry: gtk.Entry
        """
        name = entry.get_text()
        if name in self._sizegroups:
            self._add_button.set_sensitive(False)
        else:
            self._add_button.set_sensitive(True)

class CommandAddRemoveSizeGroupGadgets(FlipFlopCommandMixin, Command):
    """
    Command for adding and removing sizegroup gadgets. When adding
    gadgets to an empty sizegroup the sizegroup will be added to the
    project. When the last gadgets is removed from a sizegroup the
    sizegroup will be removed as well.
    """

    def __init__(self, sizegroup, gadgets, project, add):
        """
        Initialize the command.

        @param sizegroup: the sizegroup that the gadgets belong to
        @type sizegroup: gazpacho.sizegroup.GSizeGroup
        @param gadgets: the gadgets that should be added or removed
        @type gadgets: list (of gazpacho.gadget.Gadget)
        @param project: the project that the sizegroup belongs to
        @type project: gazpacho.project.Project
        @param add: True if the gadgets should be added
        @type add: bool
        """
        FlipFlopCommandMixin.__init__(self, add)
        if add:
            dsc = _("Add widgets to size group '%s'") % sizegroup.name
        else:
            dsc = _("Remove widgets from size group '%s'") % sizegroup.name


        Command.__init__(self, dsc)

        self._sizegroup = sizegroup
        self._gadgets = gadgets
        self._project = project

    def _execute_add(self):
        """
        Add the gadgets to the sizegroup. This might mean that the
        sizegroup will be added as well.
        """
        if self._sizegroup.is_empty():
            self._project.add_sizegroup(self._sizegroup)

        self._sizegroup.add_gadgets(self._gadgets)
    _execute_state1 = _execute_add

    def _execute_remove(self):
        """
        Remove the gadgets from the sizegroup. This might cause the
        sizegroup to be removed as well.
        """
        self._sizegroup.remove_gadgets(self._gadgets)

        if self._sizegroup.is_empty():
            self._project.remove_sizegroup(self._sizegroup)
    _execute_state2 = _execute_remove



from django import template
from django.conf import settings
from django.core import urlresolvers
from django.utils.safestring import mark_safe
from l10n.l10n_settings import get_l10n_setting
from product.models import Category
from satchmo_utils.numbers import trunc_decimal
from satchmo_utils.json import json_encode
from threaded_multihost import threadlocals
import logging
import math

log = logging.getLogger('shop.templatetags.satchmo_util')

register = template.Library()

def debug_mode(value):
    """Return true if site is in debug mode"""
    if settings.DEBUG:
        return "true"
    return ""

register.filter('debug_mode', debug_mode)

def template_range(value):
    """Return a range 1..value"""
    try:
        value = int(value)
    except:
        value = 0
    return range(1, value + 1)

register.filter('template_range', template_range)

def in_list(value, val=None):
    """returns "true" if the value is in the list"""
    if val in value:
        return "true"
    return ""

register.filter('in_list', in_list)

def as_json(value):
    """Return the value as a json encoded object"""
    return mark_safe(json_encode(value))

register.filter('as_json', as_json)

def blackbird_logging(context):
    form = context.get('form', None)
    return {
        'debug' : settings.DEBUG,
        'media_url' : settings.MEDIA_URL,
        'form' : form,
        }

register.inclusion_tag('shop/_blackbird_logging.html', takes_context=True)(blackbird_logging)

def truncate_decimal(val, places=2):
    return trunc_decimal(val, places)

register.filter('truncate_decimal', truncate_decimal)

def tag_attr(obj, arg1):
    att, value = arg1.split("=")
    obj.field.widget.attrs[att] = value
    return obj

register.filter('tag_attr', tag_attr)

def shuffle(l):
    """
    Returns the shuffled list.
    """
    import random
    l = list(l)
    random.shuffle(l)
    return l

register.filter('shuffle', shuffle)

def remove_tags(value):
    """
    Returns the text with all tags removed
    This can fail if give invalid input. This is only intended to be used on safe HTML markup. It should not be used to
    clean unsafe input data.
    For example, this will fail.

    Example::

        >> remove_tags('<<> <test></test>')
            '<test></test>'
    """
    i = value.find('<')
    last = -1
    out = []
    if i == -1:
        return value

    while i>-1:
        out.append(value[last+1:i])
        last = value.find(">", i)
        if last > -1:
            i = value.find("<", last)
        else:
            break

    if last > -1:
        out.append(value[last+1:])

    ret = " ".join(out)
    ret = ret.replace("  ", " ")
    ret = ret.replace("  ", " ")
    if ret.endswith(" "):
        ret = ret[:-1]
    return ret

register.filter('remove_tags', remove_tags)

def lookup(value, key):
    """
    Return a dictionary lookup of key in value
    """
    try:
        return value[key]
    except KeyError:
        return ""

register.filter('lookup', lookup)

def is_mod(value, args=""):
    try:
        val = int(value)
        mod = int(args)
        if val%mod == 0:
            return "true"
    except:
        pass

    return ""

register.filter('is_mod', is_mod)

def more_than(value, args=""):
    try:
        val = int(value)
        more = int(args)
        if val > more:
            return "true"
    except:
        pass

    return ""

register.filter('more_than', more_than)

def satchmo_category_search_form(category=None):
    """
    Display the form for customer to specify category to search.
    """
    try:
        url = urlresolvers.reverse('satchmo_search')
    except urlresolvers.NoReverseMatch:
        url = ""
        log.warning('No url found for satchmo_search (OK if running tests)')

    cats = Category.objects.root_categories()
    return {
        'satchmo_search_url' : url,
        'categories' : cats,
        'category' : category,
    }
register.inclusion_tag("shop/_search.html", takes_context=False)(satchmo_category_search_form)

def satchmo_language_selection_form(context):
    """
    Display the set language form, if enabled in shop settings.
    """
    request = threadlocals.get_current_request()
    enabled = get_l10n_setting('allow_translation_choice')
    languages = []
    if enabled:
        try:
            url = urlresolvers.reverse('satchmo_set_language')
            languages = settings.LANGUAGES

        except urlresolvers.NoReverseMatch:
            url = ""
            log.warning('No url found for satchmo_set_language (OK if running tests)')

    else:
        url = ""

    media_url = context.get('media_url', None)
    return {
        'enabled' : enabled,
        'set_language_url' : url,
        'languages' : languages,
        'media_url' : media_url,
        'django_language' : request.session.get('django_language', 'en'),
    }
register.inclusion_tag("l10n/_language_selection_form.html", takes_context=True)(satchmo_language_selection_form)

def satchmo_search_form():
    """
    Display the search form.
    """
    try:
        url = urlresolvers.reverse('satchmo_search')
    except urlresolvers.NoReverseMatch:
        url = ""
        log.warning('No url found for satchmo_search (OK if running tests)')

    return {
        'satchmo_search_url' : url,
        'categories' : None,
    }
register.inclusion_tag("shop/_search.html", takes_context=False)(satchmo_search_form)

def pounds(weight):
    """
    Finds the weight of a cart item, taking into consideration the quantity in
    the order.
    """
    return int(weight)
register.filter('pounds', pounds)

def ounces(weight):
    fract = weight - pounds(weight)
    return int(math.ceil(fract * 16))
register.filter('ounces', ounces)



# This is the core test suite to port Django urls.py to Routes. YMMV
from test_support import unittest
from notmm.utils.urlmap import *

class RegexURLMapTestCase(unittest.TestCase):
    
    def setUp(self):
        # default behavior 
        self.urlmap = RegexURLMap()
    
    def tearDown(self):
        #self.urlmap.__del__()
        del self.urlmap

    #def test_label(self):
    #    # verifies the default label
    #    self.failUnlessEqual(repr(self.urlmap) == "<RegexURLMap: 'yadayada'>", repr(self.urlmap))

    def test_add_routes(self):
        urlmap = self.urlmap
        
        # XXX: Should we raise an RegexURLMapException here, as 'foobar' matches
        # nothing known ?
        urlmap.add_routes('sandbox.views', 
            url(r'^$', 'foobar', name="foobar"),
            url(r'^index.php$', 'index', name="index_php"),
            url(r'^test_db_session', 'test_session', name="test_session")
            )
        # Test adding another set of routes
        urlmap.add_routes('sandbox.handlers',
            url(r'^$', 'handle500', name="index2") # XXX: Should we raise
                                                   # RegexURLException here
                                                   # too, since we're creating
                                                   # a conflict. urlmap.routes
                                                   # must be a sequence of
                                                   # unique routes.
            )
        self.assertEqual(len(urlmap)==urlmap.__len__(), True)
    
    def test_include(self):
        # default string based include
        urlmap = RegexURLMap()
        urlmap.include('sandbox.config.urls')
        self.assertEqual(len(urlmap)==len(urlmap.routes), True, len(urlmap))
    
    #def test_autodiscover(self):
    #    # test the django admin configuration (autodiscovery of admin views)
    #    from django.contrib import admin
    #    admin.autodiscover()
    #    urlmap = RegexURLMap()
    #    urlmap.include(admin.site.urls, prefix='^admin/')
    #    self.assertEqual(len(urlmap.routes)>= 11, True)
        
    def test_commit(self):
        # raise an exception
        self.urlmap.include('sandbox.config.extras')
        self.urlmap.commit()
        self.failUnlessRaises(RegexURLMapException, self.urlmap.include, ('sandbox.config.urls'))




'''
DDL statements for Elixir.

Entities having the perform_ddl statement, will automatically execute the
given DDL statement, at the given moment: ether before or after the table
creation in SQL.

The 'when' argument can be either 'before-create' or 'after-create'.
The 'statement' argument can be one of:

- a single string statement
- a list of string statements, in which case, each of them will be executed
  in turn.
- a callable which should take no argument and return either a single string
  or a list of strings.

In each string statement, you may use the special '%(fullname)s' construct,
that will be replaced with the real table name including schema, if unknown
to you. Also, self explained '%(table)s' and '%(schema)s' may be used here.

You would use this extension to handle non elixir sql statemts, like triggers
etc.

.. sourcecode:: python

    class Movie(Entity):
        title = Field(Unicode(30), primary_key=True)
        year = Field(Integer)

        perform_ddl('after-create',
                    "insert into %(fullname)s values ('Alien', 1979)")

preload_data is a more specific statement meant to preload data in your
entity table from a list of tuples (of fields values for each row).

.. sourcecode:: python

    class Movie(Entity):
        title = Field(Unicode(30), primary_key=True)
        year = Field(Integer)

        preload_data(('title', 'year'),
                     [(u'Alien', 1979), (u'Star Wars', 1977)])
        preload_data(('year', 'title'),
                     [(1982, u'Blade Runner')])
        preload_data(data=[(u'Batman', 1966)])
'''

from elixir.statements import Statement
from elixir.properties import EntityBuilder
from sqlalchemy import DDL

__all__ = ['perform_ddl', 'preload_data']
__doc_all__ = []

#
# the perform_ddl statement
#
class PerformDDLEntityBuilder(EntityBuilder):

    def __init__(self, entity, when, statement, on=None, context=None):
        self.entity = entity
        self.when = when
        self.statement = statement
        self.on = on
        self.context = context

    def after_table(self):
        statement = self.statement
        if hasattr(statement, '__call__'):
            statement = statement()
        if not isinstance(statement, list):
            statement = [statement]
        for s in statement:
            ddl = DDL(s, self.on, self.context)
            ddl.execute_at(self.when, self.entity.table)

perform_ddl = Statement(PerformDDLEntityBuilder)

#
# the preload_data statement
#
class PreloadDataEntityBuilder(EntityBuilder):

    def __init__(self, entity, columns=None, data=None):
        self.entity = entity
        self.columns = columns
        self.data = data

    def after_table(self):
        all_columns = [col.name for col in self.entity.table.columns]
        def onload(event, schema_item, connection):
            columns = self.columns
            if columns is None:
                columns = all_columns
            data = self.data
            if hasattr(data, '__call__'):
                data = data()
            insert = schema_item.insert()
            connection.execute(insert,
                [dict(zip(columns, values)) for values in data])

        self.entity.table.append_ddl_listener('after-create', onload)

preload_data = Statement(PreloadDataEntityBuilder)

